# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Caching strategies for stencil generation."""

import abc
import inspect
import pathlib
import pickle
import sys
import types
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from cached_property import cached_property

from gt4py.cartesian import config as gt_config
from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.definitions import StencilID


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_builder import StencilBuilder


class CachingStrategy(abc.ABC):
    name: str

    def __init__(self, builder: "StencilBuilder"):
        self.builder = builder

    @property
    @abc.abstractmethod
    def root_path(self) -> pathlib.Path:
        """Calculate and if necessary create the root path for stencil generation operations."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def backend_root_path(self) -> pathlib.Path:
        """
        Calculate and if necessary create the caching base path for the backend being used.

        Avoid trying to create the root_path automatically.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def cache_info_path(self) -> Optional[pathlib.Path]:
        """Calculate the file path where caching info for the current process should be stored."""
        raise NotImplementedError

    @abc.abstractmethod
    def generate_cache_info(self) -> Dict[str, Any]:
        """
        Generate the cache info dict.

        Backend specific additions can be added via a hook propery on the backend instance.
        Override :py:meth:`gt4py.backend.base.Backend.extra_cache_info` to store extra
        info.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update_cache_info(self) -> None:
        """
        Regenerate cache info and update the cache info file.

        Parameters
        ----------
        extra_cache_info:
            Additional info to be written

        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_cache_info_available_and_consistent(self, *, validate_hash: bool) -> bool:
        """
        Check if the cache can be read and is consistent.

        If the cache info file can be read, try to validate it against the current info
        as returned by py:meth:`generate_cache_info`.

        Parameters
        ----------
        validate_hash:
            Should the stencil fingerprint and source code hash be validated or not
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def stencil_id(self) -> StencilID:
        """Calculate the stencil ID used for naming and cache consistency checks."""
        pass

    @property
    @abc.abstractmethod
    def cache_info(self) -> Dict[str, Any]:
        """
        Read currently stored cache info from file into a dictionary.

        Empty if file is not found or not readable or no caching is intended.
        """
        pass

    @property
    def module_prefix(self) -> str:
        """
        Generate the prefix for stencil modules.

        Backends use this property to determine the stencil module filename.
        This may help distinguish python modules from other files.

        The default is an empty string and is suitable for situations where no caching
        is required and the generated module is intended to be imported manually.
        """
        return ""

    @property
    def module_postfix(self) -> str:
        """
        Generate the postfix for stencil modules, can depend on the fingerprint.

        Backends use this property when generating the module filename.

        .. code-block: python
            :title: Example

            @property
            def module_postfix(self) -> str:
                return f"_{self.stencil_id.version}"

        The example code would cause the generated module and class names to depend on the
        fingerprint of the stencil. That would ensure that after changing and regenerating a
        stencil it's module can be loaded without unloading the previous version first.  The
        default is empty and is suitable for when no caching is used.

        Can also be re-used elsewhere for similar purposes.
        """
        return ""

    @property
    def class_name(self) -> str:
        """Calculate the name for the stencil class, default is to read from build options."""
        return self.builder.options.name

    def capture_externals(self) -> Dict[str, Any]:
        """Extract externals from the annotated stencil definition for fingerprinting. Freezes the references."""
        return {}


class JITCachingStrategy(CachingStrategy):
    """
    Caching strategy for JIT stencil generation.

    Assign a fingerprint to the stencil being generated based on the definition function
    and some of the build options. During generation, store this fingerprint along with
    other information in a cache info file.

    In order to decide whether the stencil can be loaded from cache, check if a cache info file
    exists in the location corresponding to the current stencil. If it exists, compare it to
    the additional caching information for the current stencil. If the cache is consistent, a
    rebuild can be avoided.
    """

    name = "jit"

    _root_path: str
    _dir_name: str

    def __init__(
        self,
        builder: "StencilBuilder",
        *,
        root_path: Optional[str] = None,
        dir_name: Optional[str] = None,
    ):
        super().__init__(builder)
        self._root_path = root_path or gt_config.cache_settings["root_path"]
        self._dir_name = dir_name or gt_config.cache_settings["dir_name"]

    @property
    def root_path(self) -> pathlib.Path:
        cache_root = pathlib.Path(self._root_path) / self._dir_name

        if not cache_root.exists():
            gt_utils.make_dir(str(cache_root), is_cache=True)

        return cache_root

    @property
    def backend_root_path(self) -> pathlib.Path:
        cpython_id = "py{version.major}{version.minor}_{api_version}".format(
            version=sys.version_info, api_version=sys.api_version
        )
        backend_root = self.root_path / cpython_id / gt_utils.slugify(self.builder.backend.name)
        if not backend_root.exists():
            if not backend_root.parent.exists():
                backend_root.parent.mkdir(parents=False, exist_ok=True)
            backend_root.mkdir(parents=False, exist_ok=True)
        return backend_root

    @property
    def cache_info_path(self) -> Optional[pathlib.Path]:
        """Get the cache info file path from the stencil module path."""
        return self.builder.module_path.parent / f"{self.builder.module_path.stem}.cacheinfo"

    def generate_cache_info(self) -> Dict[str, Any]:
        return {
            "backend": self.builder.backend.name,
            "stencil_name": self.builder.stencil_id.qualified_name,
            "stencil_version": self.builder.stencil_id.version,
            "module_shash": gt_utils.shash(self.builder.stencil_source),
            **self.builder.backend.extra_cache_info,
        }

    def update_cache_info(self) -> None:
        if not self.cache_info_path:
            return
        cache_info = self.generate_cache_info()
        self.cache_info_path.parent.mkdir(parents=True, exist_ok=True)
        with self.cache_info_path.open("wb") as cache_info_file:
            pickle.dump(cache_info, cache_info_file)

    def is_cache_info_available_and_consistent(
        self, *, validate_hash: bool, catch_exceptions: bool = True
    ) -> bool:
        result = True
        if not self.cache_info_path and catch_exceptions:
            return False
        try:
            cache_info = self.generate_cache_info()
            cache_info_ns = types.SimpleNamespace(**cache_info)
            validate_extra = {
                k: v
                for k, v in self.builder.backend.extra_cache_info.items()
                if k in self.builder.backend.extra_cache_validation_keys
            }
            source = self.builder.module_path.read_text()
            module_shash = gt_utils.shash(source)

            if validate_hash:
                result = (
                    cache_info_ns.backend == self.builder.backend.name
                    and cache_info_ns.stencil_name == self.stencil_id.qualified_name
                    and cache_info_ns.stencil_version == self.stencil_id.version
                    and cache_info_ns.module_shash == module_shash
                )
                if validate_extra:
                    result &= all(
                        [cache_info[key] == validate_extra[key] for key in validate_extra]
                    )
        except Exception as err:
            if not catch_exceptions:
                raise err
            result = False

        return result

    @property
    def cache_info(self) -> Dict[str, Any]:
        if not self.cache_info_path:
            return {}
        if not self.cache_info_path.exists():
            return {}
        return self._unpickle_cache_info_file(self.cache_info_path)

    @staticmethod
    def _unpickle_cache_info_file(cache_info_path: pathlib.Path) -> Dict[str, Any]:
        with cache_info_path.open("rb") as cache_info_file:
            return pickle.load(cache_info_file)

    @property
    def options_id(self) -> str:
        if hasattr(self.builder.backend, "filter_options_for_id"):
            return self.builder.backend.filter_options_for_id(self.builder.options).shashed_id
        return self.builder.options.shashed_id

    def capture_externals(self) -> Dict[str, Any]:
        """Extract externals from the annotated stencil definition for fingerprinting."""
        return self._externals

    @cached_property
    def _externals(self) -> Dict[str, Any]:
        """Extract externals from the annotated stencil definition for fingerprinting."""
        return {
            name: value._gtscript_["canonical_ast"] if hasattr(value, "_gtscript_") else value
            for name, value in self.builder.definition._gtscript_["externals"].items()
        }

    def _extract_api_annotations(self) -> List[str]:
        """Extract API annotations from the annotated stencil definition for fingerprinting."""
        return [str(item) for item in self.builder.definition._gtscript_["api_annotations"]]

    @property
    def stencil_id(self) -> StencilID:
        fingerprint = {
            "__main__": self.builder.definition._gtscript_["canonical_ast"],
            "docstring": inspect.getdoc(self.builder.definition),
            "api_annotations": f"[{', '.join(self._extract_api_annotations())}]",
            **self._externals,
        }

        # typeignore because attrclass StencilID has generated constructor
        return StencilID(  # type: ignore
            self.builder.options.qualified_name,
            gt_utils.shashed_id(gt_utils.shashed_id(fingerprint), self.options_id),
        )

    @property
    def module_prefix(self) -> str:
        return "m_"

    @property
    def module_postfix(self) -> str:
        backend_name = gt_utils.slugify(self.builder.backend.name)
        id_version = self.stencil_id.version
        return f"__{backend_name}_{id_version}"

    @property
    def class_name(self) -> str:
        name = self.builder.options.name
        return f"{name}__{self.module_postfix}"


class NoCachingStrategy(CachingStrategy):
    """
    Apply no caching, useful for CLI.

    Instead of calculating paths and names based on a stencil fingerprint, use
    user input or by default the current working dir to output source and build files.

    This caching strategy is incapable of avoiding rebuilds and after rebuilding at runtime
    the old stencil and it's module has to be unloaded by the user before reloading for any
    changes to take effect.

    All methods related to generating or checking the cache info are noops and cache validation
    always fails.

    Parameters
    ----------
    builder:
        A stencil builder instance

    output_path:
        The path to where all source code should be written
    """

    name = "nocaching"

    def __init__(self, builder: "StencilBuilder", *, output_path: pathlib.Path = pathlib.Path(".")):
        super().__init__(builder)
        self._output_path = output_path

    @property
    def root_path(self) -> pathlib.Path:
        """Get stencil code output path set during initialization."""
        if not self._output_path.exists():
            self._output_path.mkdir(parents=True)
        return self._output_path

    @property
    def backend_root_path(self) -> pathlib.Path:
        """Simply pass through the user chosen output path."""
        return self.root_path

    @property
    def cache_info_path(self) -> Optional[pathlib.Path]:
        return None

    def generate_cache_info(self) -> Dict[str, Any]:
        return {}

    def update_cache_info(self) -> None:
        pass

    def is_cache_info_available_and_consistent(self, *, validate_hash: bool) -> bool:
        return False

    @property
    def cache_info(self) -> Dict[str, Any]:
        return {}

    @property
    def stencil_id(self) -> StencilID:
        """Get a fingerprint-less stencil id."""
        # ignore type because mypy does not understand attrib classes
        return StencilID(  # type: ignore
            qualified_name=self.builder.options.qualified_name, version=""
        )


def strategy_factory(
    name: str, builder: "StencilBuilder", *args: Any, **kwargs: Any
) -> CachingStrategy:
    strategies = {"jit": JITCachingStrategy, "nocaching": NoCachingStrategy}
    return strategies[name](builder, *args, **kwargs)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Package metadata: version, authors, license and copyright."""

from typing import Final

from packaging import version as pkg_version


__all__ = [
    "__author__",
    "__copyright__",
    "__license__",
    "__version__",
    "__version_info__",
]


__author__: Final = "ETH Zurich and individual contributors"
__copyright__: Final = "Copyright (c) 2014-2022 ETH Zurich"
__license__: Final = "GPL-3.0-or-later"


__version__: Final = "1.0.0"
__version_info__: Final = pkg_version.parse(__version__)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections import defaultdict
from typing import Any, Dict

from gt4py.eve import NodeTranslator

from . import cuir


class CacheExtents(NodeTranslator):
    def visit_IJCacheDecl(
        self,
        node: cuir.IJCacheDecl,
        *,
        ij_extents: Dict[str, cuir.KExtent],
        **kwargs: Any,
    ) -> cuir.IJCacheDecl:
        return cuir.IJCacheDecl(name=node.name, dtype=node.dtype, extent=ij_extents[node.name])

    def visit_KCacheDecl(
        self, node: cuir.KCacheDecl, *, k_extents: Dict[str, cuir.KExtent], **kwargs: Any
    ) -> cuir.KCacheDecl:
        return cuir.KCacheDecl(name=node.name, dtype=node.dtype, extent=k_extents[node.name])

    def visit_VerticalLoop(self, node: cuir.VerticalLoop) -> cuir.VerticalLoop:
        ij_extents: Dict[str, cuir.IJExtent] = defaultdict(cuir.IJExtent.zero)
        for horizontal_execution in node.walk_values().if_isinstance(cuir.HorizontalExecution):
            ij_access_extents = (
                horizontal_execution.walk_values()
                .if_isinstance(cuir.IJCacheAccess)
                .reduceby(
                    lambda acc, x: acc.union(cuir.IJExtent.from_offset(x.offset)),
                    "name",
                    init=cuir.IJExtent.zero(),
                )
            )
            for field, ij_access_extent in ij_access_extents:
                ij_extents[field] = ij_extents[field].union(
                    ij_access_extent + horizontal_execution.extent
                )

        k_extents = (
            node.walk_values()
            .if_isinstance(cuir.KCacheAccess)
            .reduceby(
                lambda acc, x: acc.union(cuir.KExtent.from_offset(x.offset)),
                "name",
                init=cuir.KExtent.zero(),
                as_dict=True,
            )
        )

        return self.generic_visit(node, ij_extents=ij_extents, k_extents=k_extents)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import copy
from typing import TYPE_CHECKING, Dict, List

import dace
import dace.data
import dace.library
import dace.subsets
import sympy

from gt4py.cartesian.gtc import daceir as dcir
from gt4py.cartesian.gtc.dace.expansion.daceir_builder import DaCeIRBuilder
from gt4py.cartesian.gtc.dace.expansion.sdfg_builder import StencilComputationSDFGBuilder

from .utils import split_horizontal_executions_regions


if TYPE_CHECKING:
    from gt4py.cartesian.gtc.dace.nodes import StencilComputation


class StencilComputationExpansion(dace.library.ExpandTransformation):
    environments: List = []

    @staticmethod
    def _solve_for_domain(field_decls: Dict[str, dcir.FieldDecl], outer_subsets):
        equations = []
        symbols = set()

        # Collect equations and symbols from arguments and shapes
        for field, decl in field_decls.items():
            inner_shape = [dace.symbolic.pystr_to_symbolic(s) for s in decl.shape]
            outer_shape = [
                dace.symbolic.pystr_to_symbolic(s) for s in outer_subsets[field].bounding_box_size()
            ]

            for inner_dim, outer_dim in zip(inner_shape, outer_shape):
                repldict = {}
                for sym in dace.symbolic.symlist(inner_dim).values():
                    newsym = dace.symbolic.symbol("__SOLVE_" + str(sym))
                    symbols.add(newsym)
                    repldict[sym] = newsym

                # Replace symbols with __SOLVE_ symbols so as to allow
                # the same symbol in the called SDFG
                if repldict:
                    inner_dim = inner_dim.subs(repldict)

                equations.append(inner_dim - outer_dim)
        if len(symbols) == 0:
            return {}

        # Solve for all at once
        results = sympy.solve(equations, *symbols, dict=True)
        result = results[0]
        result = {str(k)[len("__SOLVE_") :]: v for k, v in result.items()}
        return result

    @staticmethod
    def _fix_context(
        nsdfg, node: "StencilComputation", parent_state: dace.SDFGState, daceir: dcir.NestedSDFG
    ):
        """Apply changes to StencilComputation and the SDFG it is embedded in to satisfy post-expansion constraints.

        * change connector names to match inner array name (before expansion prefixed to satisfy uniqueness)
        * change in- and out-edges' subsets so that they have the same shape as the corresponding array inside
        * determine the domain size based on edges to StencilComputation

        """
        # change connector names
        for in_edge in parent_state.in_edges(node):
            assert in_edge.dst_conn.startswith("__in_")
            in_edge.dst_conn = in_edge.dst_conn[len("__in_") :]
        for out_edge in parent_state.out_edges(node):
            assert out_edge.src_conn.startswith("__out_")
            out_edge.src_conn = out_edge.src_conn[len("__out_") :]

        # union input and output subsets
        subsets = {}
        for edge in parent_state.in_edges(node):
            subsets[edge.dst_conn] = edge.data.subset
        for edge in parent_state.out_edges(node):
            subsets[edge.src_conn] = dace.subsets.union(
                edge.data.subset, subsets.get(edge.src_conn, edge.data.subset)
            )
        # ensure single-use of input and output subset instances
        for edge in parent_state.in_edges(node):
            edge.data.subset = copy.deepcopy(subsets[edge.dst_conn])
        for edge in parent_state.out_edges(node):
            edge.data.subset = copy.deepcopy(subsets[edge.src_conn])

        # determine "__I", "__J" and "__K" values based on memlets to StencilComputation's shape
        symbol_mapping = StencilComputationExpansion._solve_for_domain(
            {
                decl.name: decl
                for decl in daceir.field_decls
                if decl.name
                in set(memlet.field for memlet in daceir.read_memlets + daceir.write_memlets)
            },
            subsets,
        )
        nsdfg.symbol_mapping.update({**symbol_mapping, **node.symbol_mapping})

        # remove unused symbols from symbol_mapping
        delkeys = set()
        for sym in node.symbol_mapping.keys():
            if str(sym) not in nsdfg.sdfg.free_symbols:
                delkeys.add(str(sym))
        for key in delkeys:
            del node.symbol_mapping[key]
            if key in nsdfg.symbol_mapping:
                del nsdfg.symbol_mapping[key]

    @staticmethod
    def _get_parent_arrays(
        node: "StencilComputation", parent_state: dace.SDFGState, parent_sdfg: dace.SDFG
    ) -> Dict[str, dace.data.Data]:
        parent_arrays: Dict[str, dace.data.Data] = {}
        for edge in (e for e in parent_state.in_edges(node) if e.dst_conn is not None):
            parent_arrays[edge.dst_conn[len("__in_") :]] = parent_sdfg.arrays[edge.data.data]
        for edge in (e for e in parent_state.out_edges(node) if e.src_conn is not None):
            parent_arrays[edge.src_conn[len("__out_") :]] = parent_sdfg.arrays[edge.data.data]
        return parent_arrays

    @staticmethod
    def expansion(
        node: "StencilComputation", parent_state: dace.SDFGState, parent_sdfg: dace.SDFG
    ) -> dace.nodes.NestedSDFG:
        """Expand the coarse SDFG in parent_sdfg to a NestedSDFG with all the states."""
        split_horizontal_executions_regions(node)
        arrays = StencilComputationExpansion._get_parent_arrays(node, parent_state, parent_sdfg)

        daceir: dcir.NestedSDFG = DaCeIRBuilder().visit(
            node.oir_node, global_ctx=DaCeIRBuilder.GlobalContext(library_node=node, arrays=arrays)
        )

        nsdfg: dace.nodes.NestedSDFG = StencilComputationSDFGBuilder().visit(daceir)

        StencilComputationExpansion._fix_context(nsdfg, node, parent_state, daceir)
        return nsdfg



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Basic utilities for Python programming."""

import collections.abc
import functools
import hashlib
import importlib.util
import itertools
import json
import os
import string
import sys
import time
import types

from gt4py.cartesian import config as gt_config


NOTHING = object()


def slugify(value: str, *, replace_spaces=True, valid_symbols="-_.()", invalid_marker=""):
    valid_chars = valid_symbols + string.ascii_letters + string.digits
    slug = "".join(c if c in valid_chars else invalid_marker for c in value)
    if replace_spaces:
        slug = slug.replace(" ", "_")
    return slug


# def stringify(value):
#     pass


def jsonify(value, indent=2):
    return json.dumps(value, indent=indent, default=lambda obj: str(obj))


def is_identifier_name(value, namespaced=True):
    if isinstance(value, str):
        if namespaced:
            return all(name.isidentifier() for name in value.split("."))
        else:
            return value.isidentifier()
    else:
        return False


def flatten(nested_iterables, filter_none=False, *, skip_types=(str, bytes)):
    return list(flatten_iter(nested_iterables, filter_none, skip_types=skip_types))


def flatten_iter(nested_iterables, filter_none=False, *, skip_types=(str, bytes)):
    for item in nested_iterables:
        if isinstance(item, collections.abc.Iterable) and not isinstance(item, skip_types):
            yield from flatten(item, filter_none)
        else:
            if item is not None or not filter_none:
                yield item


def get_member(instance, item_name):
    try:
        if (
            isinstance(instance, collections.abc.Mapping)
            or isinstance(instance, collections.abc.Sequence)
            and isinstance(item_name, int)
        ):
            return instance[item_name]
        else:
            return getattr(instance, item_name)
    except Exception:
        return NOTHING


def compose(*functions_or_iterable):
    """Return a function that chains the input functions.

    Derived from: https://mathieularose.com/function-composition-in-python/

    """
    if len(functions_or_iterable) == 1 and isinstance(
        functions_or_iterable[0], collections.abc.Iterable
    ):
        functions = functions_or_iterable[0]
    else:
        functions = functions_or_iterable

    return functools.reduce(
        lambda this_func, previous_func: lambda x: this_func(previous_func(x)),
        functions,
        lambda x: x,
    )


def is_collection(value, *, optional=False, accept_mapping=True):
    if value is None and optional:
        return True
    if (
        value is not None
        and isinstance(value, collections.abc.Iterable)
        and not isinstance(value, (str, bytes))
    ):
        return accept_mapping or not isinstance(value, collections.abc.Mapping)

    return False


def is_iterable_of(
    value,
    item_class,
    iterable_class=collections.abc.Iterable,
    *,
    optional=False,
    accept_mapping=True,
):
    try:
        if value is None and optional:
            return True
        if value is not None and (
            isinstance(value, iterable_class) and all([isinstance(i, item_class) for i in value])
        ):
            return accept_mapping or not isinstance(value, collections.abc.Mapping)

    except Exception:
        pass

    return False


def is_mapping_of(
    value, key_class, value_class=None, mapping_class=collections.abc.Mapping, *, optional=False
):
    try:
        if value is None and optional:
            return True
        if (
            value is not None
            and (isinstance(value, mapping_class))
            and all([isinstance(k, key_class) for k in value.keys()])
            and (value_class is None or all([isinstance(v, value_class) for v in value.values()]))
        ):
            return True

    except Exception:
        pass

    return False


def normalize_mapping(mapping, key_types=(object,), *, filter_none=False):
    assert isinstance(mapping, collections.abc.Mapping)
    if isinstance(key_types, collections.abc.Iterable):
        assert all(isinstance(k, type) for k in key_types)
        key_types = tuple(key_types)
    result = {}
    for key, value in mapping.items():
        if not filter_none or value is not None:
            if is_iterable_of(key, key_types, iterable_class=(tuple, list)):
                for k in key:
                    result[k] = value
            else:
                result[key] = value

    return result


def shash(*args, hash_algorithm=None):
    if hash_algorithm is None:
        hash_algorithm = hashlib.sha256()

    for item in args:
        if not isinstance(item, bytes):
            if isinstance(item, collections.abc.Mapping):
                item = list(flatten(sorted(item.items())))
            elif not isinstance(item, str) and isinstance(item, collections.abc.Iterable):
                item = list(flatten(item))
            item = str.encode(repr(item))
        hash_algorithm.update(item)

    return hash_algorithm.hexdigest()


def shashed_id(*args, length=10, hash_algorithm=None):
    return shash(*args, hash_algorithm=hash_algorithm)[:length]


def classmethod_to_function(class_method, instance=None, owner=type(None), remove_cls_arg=False):
    if remove_cls_arg:
        return functools.partial(class_method.__get__(instance, owner), None)
    else:
        return class_method.__get__(instance, owner)


def namespace_from_nested_dict(nested_dict):
    assert isinstance(nested_dict, dict)
    return types.SimpleNamespace(
        **{
            key: namespace_from_nested_dict(value) if isinstance(value, dict) else value
            for key, value in nested_dict.items()
        }
    )


def make_local_dir(dir_name, base_dir=None, *, mode=0o777, is_package=False, is_cache=False):
    if base_dir is None:
        base_dir = os.getcwd()
    dir_name = os.path.join(base_dir, dir_name)
    return make_dir(dir_name)


def make_dir(dir_name, *, mode=0o777, is_package=False, is_cache=False):
    dir_name = os.path.abspath(dir_name)
    os.makedirs(dir_name, mode=mode, exist_ok=True)

    if is_package:
        with open(os.path.join(dir_name, "__init__.py"), "a"):
            pass

    if is_cache:
        with open(os.path.join(dir_name, "CACHEDIR.TAG"), "w") as f:
            f.write(
                """Signature: 8
a477f597d28d172789f06886806bc55
# This file is a cache directory tag created by GT4Py.
# For information about cache directory tags, see:
#	http://www.brynosaurus.com/cachedir/
"""
            )

    return dir_name


def make_module_from_file(qualified_name, file_path, *, public_import=False):
    """Import module from file.

    References:
      https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
      https://stackoverflow.com/a/43602645

    """

    def load():
        spec = importlib.util.spec_from_file_location(qualified_name, file_path)
        if not spec:
            raise ModuleNotFoundError(f"No module named '{qualified_name}'")
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if public_import:
            sys.modules[qualified_name] = module
            package_name = getattr(module, "__package__", "")
            if not package_name:
                package_name = ".".join(qualified_name.split(".")[:-1])
                module.__package__ = package_name
            components = package_name.split(".")
            module_name = qualified_name.split(".")[-1]

            if components[0] in sys.modules:
                parent = sys.modules[components[0]]
                for current in components[1:]:
                    parent = getattr(parent, current, None)
                    if not parent:
                        break
                else:
                    setattr(parent, module_name, module)
        return module

    for _i in range(max(gt_config.cache_settings["load_retries"], 0)):
        try:
            return load()
        except ModuleNotFoundError:
            time.sleep(max(gt_config.cache_settings["load_retry_delay"], 0) / 1000)

    return load()


def patch_module(module, member, new_value, *, recursive=True):
    """Monkey patch a module replacing a member with a new value."""
    if not isinstance(module, types.ModuleType):
        raise ValueError("Invalid 'module' argument")

    pending = [module]
    visited = set()
    originals = {}

    while pending:
        patched = {}
        current, pending = pending[0], pending[1:]
        visited.add(current)

        for name, value in current.__dict__.items():
            if value is member:
                patched[name] = value
                current.__dict__[name] = new_value
            elif isinstance(value, types.ModuleType) and value not in visited:
                pending.append(value)

        if patched:
            originals[current] = patched

    patch = dict(
        module=module,
        original_value=member,
        patched_value=new_value,
        recursive=recursive,
        originals=originals,
    )

    return patch


def restore_module(patch, *, verify=True):
    """Restore a module patched with the `patch_module()` function."""
    if not isinstance(patch, dict) or not {
        "module",
        "original_value",
        "patched_value",
        "recursive",
        "originals",
    } <= set(patch.keys()):
        raise ValueError("Invalid 'patch' definition")

    patched_value = patch["patched_value"]
    original_value = patch["original_value"]

    for current, changes in patch["originals"].items():
        for name, value in changes.items():
            assert (
                value is original_value and current.__dict__[name] is patched_value
            ) or not verify
            current.__dict__[name] = original_value


class Registry(dict):
    @property
    def names(self):
        return list(self.keys())

    def register(self, name, item=NOTHING):
        if name in self.keys():
            raise ValueError("Name already exists in registry")

        def _wrapper(obj):
            self[name] = obj
            return obj

        return _wrapper if item is NOTHING else _wrapper(item)


class ClassProperty:
    """Like a :class:`property`, but the wrapped function is a class method."""

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, instance, cls=None):
        if instance is None:
            if cls is None:
                return self
        elif cls is None:
            cls = type(instance)
        if self.fget is None:
            raise AttributeError("Invalid class property getter")
        return self.fget(cls)

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError("Invalid class property setter")
        self.fset(type(instance), value)

    def __delete__(self, instance):
        if self.fdel is None:
            raise AttributeError("Invalid class property deleter")
        self.fdel(type(instance))

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


classproperty = ClassProperty


class BaseFrozen:
    def __setattr__(self, key, value):
        raise AttributeError("Attempting a modification of an attribute in a frozen class")

    def __delattr__(self, item):
        raise AttributeError("Attempting a deletion of an attribute in a frozen class")


class BaseSingleton:
    def __new__(cls, *args, **kwargs):
        if getattr(cls, "_instance", None) is None:
            cls._instance = object.__new__(cls, *args, **kwargs)

        return cls._instance


class UniqueIdGenerator:
    def __init__(self, start=1):
        self.counter = itertools.count(start=start)
        self._current = None

    @property
    def new(self):
        self._current = next(self.counter)
        return self._current

    @property
    def current(self):
        return self._current



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any, Dict, Generator, List, Optional, Sequence, Set, Tuple, Union

import dace
import sympy

import gt4py.cartesian.gtc.definitions
from gt4py import eve
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.common import LocNode
from gt4py.cartesian.gtc.dace.symbol_utils import (
    get_axis_bound_dace_symbol,
    get_axis_bound_diff_str,
    get_axis_bound_str,
    get_dace_symbol,
)
from gt4py.eve import datamodels


@eve.utils.noninstantiable
class Expr(common.Expr):
    dtype: common.DataType


@eve.utils.noninstantiable
class Stmt(common.Stmt):
    pass


class Axis(eve.StrEnum):
    I = "I"  # noqa: E741 ambiguous variable name 'I'
    J = "J"
    K = "K"

    def domain_symbol(self) -> eve.SymbolRef:
        return eve.SymbolRef("__" + self.upper())

    def iteration_symbol(self) -> eve.SymbolRef:
        return eve.SymbolRef("__" + self.lower())

    def tile_symbol(self) -> eve.SymbolRef:
        return eve.SymbolRef("__tile_" + self.lower())

    @staticmethod
    def dims_3d() -> Generator["Axis", None, None]:
        yield from [Axis.I, Axis.J, Axis.K]

    @staticmethod
    def dims_horizontal() -> Generator["Axis", None, None]:
        yield from [Axis.I, Axis.J]

    def to_idx(self) -> int:
        return [Axis.I, Axis.J, Axis.K].index(self)

    def domain_dace_symbol(self):
        return get_dace_symbol(self.domain_symbol())

    def iteration_dace_symbol(self):
        return get_dace_symbol(self.iteration_symbol())

    def tile_dace_symbol(self):
        return get_dace_symbol(self.tile_symbol())


class MapSchedule(eve.IntEnum):
    Default = 0
    Sequential = 1

    CPU_Multicore = 2

    GPU_Device = 3
    GPU_ThreadBlock = 4

    def to_dace_schedule(self):
        return {
            MapSchedule.Default: dace.ScheduleType.Default,
            MapSchedule.Sequential: dace.ScheduleType.Sequential,
            MapSchedule.CPU_Multicore: dace.ScheduleType.CPU_Multicore,
            MapSchedule.GPU_Device: dace.ScheduleType.GPU_Device,
            MapSchedule.GPU_ThreadBlock: dace.ScheduleType.GPU_ThreadBlock,
        }[self]

    @classmethod
    def from_dace_schedule(cls, schedule):
        return {
            dace.ScheduleType.Default: MapSchedule.Default,
            dace.ScheduleType.Sequential: MapSchedule.Sequential,
            dace.ScheduleType.CPU_Multicore: MapSchedule.CPU_Multicore,
            dace.ScheduleType.GPU_Device: MapSchedule.GPU_Device,
            dace.ScheduleType.GPU_ThreadBlock: MapSchedule.GPU_ThreadBlock,
        }[schedule]


class StorageType(eve.IntEnum):
    Default = 0

    CPU_Heap = 1

    GPU_Global = 3
    GPU_Shared = 4

    Register = 5

    def to_dace_storage(self):
        return {
            StorageType.Default: dace.StorageType.Default,
            StorageType.CPU_Heap: dace.StorageType.CPU_Heap,
            StorageType.GPU_Global: dace.StorageType.GPU_Global,
            StorageType.GPU_Shared: dace.StorageType.GPU_Shared,
            StorageType.Register: dace.StorageType.Register,
        }[self]

    @classmethod
    def from_dace_storage(cls, schedule):
        return {
            dace.StorageType.Default: StorageType.Default,
            dace.StorageType.CPU_Heap: StorageType.CPU_Heap,
            dace.StorageType.GPU_Global: StorageType.GPU_Global,
            dace.StorageType.GPU_Shared: StorageType.GPU_Shared,
            dace.StorageType.Register: StorageType.Register,
        }[schedule]


class AxisBound(common.AxisBound):
    axis: Axis

    def __str__(self):
        return get_axis_bound_str(self, self.axis.domain_symbol())

    @classmethod
    def from_common(cls, axis, node):
        return cls(axis=axis, level=node.level, offset=node.offset)

    def to_dace_symbolic(self):
        return get_axis_bound_dace_symbol(self)


class IndexWithExtent(eve.Node):
    axis: Axis
    value: Union[AxisBound, int, str]
    extent: Tuple[int, int]

    @property
    def free_symbols(self) -> Set[eve.SymbolRef]:
        if isinstance(self.value, AxisBound) and self.value.level == common.LevelMarker.END:
            return {self.axis.domain_symbol()}
        elif isinstance(self.value, str):
            return {self.axis.iteration_symbol()}
        return set()

    @classmethod
    def from_axis(cls, axis: Axis, extent=(0, 0)):
        return cls(axis=axis, value=axis.iteration_symbol(), extent=extent)

    @property
    def size(self):
        return self.extent[1] - self.extent[0] + 1

    @property
    def overapproximated_size(self):
        return self.size

    def union(self, other: "IndexWithExtent"):
        assert self.axis == other.axis
        if isinstance(self.value, int) or (isinstance(self.value, str) and self.value.isdigit()):
            value = other.value
        elif isinstance(other.value, int) or (
            isinstance(other.value, str) and other.value.isdigit()
        ):
            value = self.value
        elif (
            self.value == self.axis.iteration_symbol()
            or other.value == self.axis.iteration_symbol()
        ):
            value = self.axis.iteration_symbol()
        else:
            assert other.value == self.value
            value = self.value
        return IndexWithExtent(
            axis=self.axis,
            value=value,
            extent=(
                min(self.extent[0], other.extent[0]),
                max(self.extent[1], other.extent[1]),
            ),
        )

    @property
    def idx_range(self):
        return (
            f"{self.value}{self.extent[0]:+d}",
            f"{self.value}{self.extent[1] + 1:+d}",
        )

    def to_dace_symbolic(self):
        if isinstance(self.value, AxisBound):
            symbolic_value = get_axis_bound_dace_symbol(self.value)
        elif isinstance(self.value, str):
            symbolic_value = next(
                axis for axis in Axis.dims_3d() if axis.iteration_symbol() == self.value
            ).iteration_dace_symbol()
        else:
            symbolic_value = self.value
        return symbolic_value + self.extent[0], symbolic_value + self.extent[1] + 1

    def shifted(self, offset):
        extent = self.extent[0] + offset, self.extent[1] + offset
        return IndexWithExtent(axis=self.axis, value=self.value, extent=extent)


class DomainInterval(eve.Node):
    start: AxisBound
    end: AxisBound

    @property
    def free_symbols(self) -> Set[eve.SymbolRef]:
        res = set()
        if self.start.level == common.LevelMarker.END:
            res.add(self.start.axis.domain_symbol())
        if self.end.level == common.LevelMarker.END:
            res.add(self.end.axis.domain_symbol())
        return res

    @property
    def size(self):
        return get_axis_bound_diff_str(
            self.end, self.start, var_name=self.start.axis.domain_symbol()
        )

    @property
    def overapproximated_size(self):
        return self.size

    @classmethod
    def union(cls, first, second):
        return cls(
            start=min(first.start, second.start),
            end=max(first.end, second.end),
        )

    @classmethod
    def intersection(cls, axis, first, second):
        first_start = first.start if first.start is not None else second.start
        first_end = first.end if first.end is not None else second.end
        second_start = second.start if second.start is not None else first.start
        second_end = second.end if second.end is not None else first_end.end

        assert (first_start <= second_end and first_end >= second_start) or (
            second_start <= first_end and second_end >= first_start
        )

        start = max(first_start, second_start)
        start = AxisBound(axis=axis, level=start.level, offset=start.offset)
        end = min(first_end, second_end)
        end = AxisBound(axis=axis, level=end.level, offset=end.offset)
        return cls(start=start, end=end)

    @property
    def idx_range(self):
        return str(self.start), str(self.end)

    def to_dace_symbolic(self):
        return self.start.to_dace_symbolic(), self.end.to_dace_symbolic()

    def shifted(self, offset: int):
        return DomainInterval(
            start=AxisBound(
                axis=self.start.axis,
                level=self.start.level,
                offset=self.start.offset + offset,
            ),
            end=AxisBound(
                axis=self.end.axis,
                level=self.end.level,
                offset=self.end.offset + offset,
            ),
        )

    def is_subset_of(self, other: "DomainInterval") -> bool:
        return self.start >= other.start and self.end <= other.end


class TileInterval(eve.Node):
    axis: Axis
    start_offset: int
    end_offset: int
    tile_size: int
    domain_limit: AxisBound

    @property
    def free_symbols(self) -> Set[eve.SymbolRef]:
        res = {
            self.axis.tile_symbol(),
        }
        if self.domain_limit.level == common.LevelMarker.END:
            res.add(self.axis.domain_symbol())
        return res

    @property
    def size(self):
        return "min({tile_size}, {domain_limit} - {tile_symbol}){halo_size:+d}".format(
            tile_size=self.tile_size,
            domain_limit=self.domain_limit,
            tile_symbol=self.axis.tile_symbol(),
            halo_size=self.end_offset - self.start_offset,
        )

    @property
    def overapproximated_size(self):
        return "{tile_size}{halo_size:+d}".format(
            tile_size=self.tile_size,
            halo_size=self.end_offset - self.start_offset,
        )

    @classmethod
    def union(cls, first, second):
        assert first.axis == second.axis
        assert first.tile_size == second.tile_size
        assert first.domain_limit == second.domain_limit
        return cls(
            axis=first.axis,
            start_offset=min(first.start_offset, second.start_offset),
            end_offset=max(first.end_offset, second.end_offset),
            tile_size=first.tile_size,
            domain_limit=first.domain_limit,
        )

    @property
    def idx_range(self):
        start = f"{self.axis.tile_symbol()}{self.start_offset:+d}"
        end = f"{start}+({self.size})"
        return start, end

    def dace_symbolic_size(self):
        return (
            sympy.Min(
                self.tile_size, self.domain_limit.to_dace_symbolic() - self.axis.tile_dace_symbol()
            )
            + self.end_offset
            - self.start_offset
        )

    def to_dace_symbolic(self):

        start = self.axis.tile_dace_symbol() + self.start_offset
        end = start + self.dace_symbolic_size()
        return start, end


class Range(eve.Node):
    var: eve.SymbolRef
    interval: Union[DomainInterval, TileInterval]
    stride: int

    @classmethod
    def from_axis_and_interval(
        cls, axis: Axis, interval: Union[DomainInterval, TileInterval], stride=1
    ):

        return cls(
            var=axis.iteration_symbol(),
            interval=interval,
            stride=stride,
        )

    @property
    def free_symbols(self) -> Set[eve.SymbolRef]:
        return {self.var, *self.interval.free_symbols}


class GridSubset(eve.Node):
    intervals: Dict[Axis, Union[DomainInterval, TileInterval, IndexWithExtent]]

    def __iter__(self):
        for axis in Axis.dims_3d():
            if axis in self.intervals:
                yield self.intervals[axis]

    def items(self):
        for axis in Axis.dims_3d():
            if axis in self.intervals:
                yield axis, self.intervals[axis]

    @property
    def free_symbols(self) -> Set[eve.SymbolRef]:
        return set().union(*(interval.free_symbols for interval in self.intervals.values()))

    @classmethod
    def single_gridpoint(cls, offset=(0, 0, 0)):
        return cls(
            intervals={
                axis: IndexWithExtent.from_axis(axis, extent=(offset[i], offset[i]))
                for i, axis in enumerate(Axis.dims_3d())
            }
        )

    @property
    def shape(self):
        return tuple(interval.size for _, interval in self.items())

    @property
    def overapproximated_shape(self):
        return tuple(interval.overapproximated_size for _, interval in self.items())

    def restricted_to_index(self, axis: Axis, extent=(0, 0)) -> "GridSubset":
        intervals = dict(self.intervals)
        intervals[axis] = IndexWithExtent.from_axis(axis, extent=extent)
        return GridSubset(intervals=intervals)

    def set_interval(
        self,
        axis: Axis,
        interval: Union[DomainInterval, IndexWithExtent, TileInterval, oir.Interval],
    ) -> "GridSubset":
        if isinstance(interval, oir.Interval):
            interval = DomainInterval(
                start=AxisBound(
                    level=interval.start.level,
                    offset=interval.start.offset,
                    axis=Axis.K,
                ),
                end=AxisBound(level=interval.end.level, offset=interval.end.offset, axis=Axis.K),
            )
        elif isinstance(interval, DomainInterval):
            assert interval.start.axis == axis
        intervals = dict(self.intervals)
        intervals[axis] = interval
        return GridSubset(intervals=intervals)

    @classmethod
    def from_gt4py_extent(cls, extent: gt4py.cartesian.gtc.definitions.Extent):
        i_interval = DomainInterval(
            start=AxisBound(level=common.LevelMarker.START, offset=extent[0][0], axis=Axis.I),
            end=AxisBound(level=common.LevelMarker.END, offset=extent[0][1], axis=Axis.I),
        )
        j_interval = DomainInterval(
            start=AxisBound(level=common.LevelMarker.START, offset=extent[1][0], axis=Axis.J),
            end=AxisBound(level=common.LevelMarker.END, offset=extent[1][1], axis=Axis.J),
        )

        return cls(intervals={Axis.I: i_interval, Axis.J: j_interval})

    @classmethod
    def from_interval(
        cls,
        interval: Union[oir.Interval, TileInterval, DomainInterval, IndexWithExtent],
        axis: Axis,
    ):
        res_interval: Union[IndexWithExtent, TileInterval, DomainInterval]
        if isinstance(interval, (DomainInterval, oir.Interval)):
            res_interval = DomainInterval(
                start=AxisBound(
                    level=interval.start.level,
                    offset=interval.start.offset,
                    axis=Axis.K,
                ),
                end=AxisBound(level=interval.end.level, offset=interval.end.offset, axis=Axis.K),
            )
        else:
            assert isinstance(interval, (TileInterval, IndexWithExtent))
            res_interval = interval

        return cls(intervals={axis: res_interval})

    def axes(self):
        for axis in Axis.dims_3d():
            if axis in self.intervals:
                yield axis

    @classmethod
    def full_domain(cls, axes=None):
        if axes is None:
            axes = Axis.dims_3d()
        res_subsets = dict()
        for axis in axes:
            res_subsets[axis] = DomainInterval(
                start=AxisBound(axis=axis, level=common.LevelMarker.START, offset=0),
                end=AxisBound(axis=axis, level=common.LevelMarker.END, offset=0),
            )
        return GridSubset(intervals=res_subsets)

    def tile(self, tile_sizes: Dict[Axis, int]):
        res_intervals: Dict[Axis, Union[DomainInterval, TileInterval, IndexWithExtent]] = {}
        for axis, interval in self.intervals.items():
            if isinstance(interval, DomainInterval) and axis in tile_sizes:
                if axis == Axis.K:
                    res_intervals[axis] = TileInterval(
                        axis=axis,
                        tile_size=tile_sizes[axis],
                        start_offset=0,
                        end_offset=0,
                        domain_limit=interval.end,
                    )
                else:
                    assert (
                        interval.start.level == common.LevelMarker.START
                        and interval.end.level == common.LevelMarker.END
                    )
                    res_intervals[axis] = TileInterval(
                        axis=axis,
                        tile_size=tile_sizes[axis],
                        start_offset=interval.start.offset,
                        end_offset=interval.end.offset,
                        domain_limit=AxisBound(axis=axis, level=common.LevelMarker.END, offset=0),
                    )
            else:
                res_intervals[axis] = interval
        return GridSubset(intervals=res_intervals)

    def union(self, other):
        assert list(self.axes()) == list(other.axes())
        intervals = dict()
        for axis in self.axes():
            interval1 = self.intervals[axis]
            interval2 = other.intervals[axis]
            if isinstance(interval1, DomainInterval) and isinstance(interval2, DomainInterval):
                intervals[axis] = DomainInterval.union(interval1, interval2)
            elif isinstance(interval1, TileInterval) and isinstance(interval2, TileInterval):
                intervals[axis] = TileInterval.union(interval1, interval2)
            elif isinstance(interval1, IndexWithExtent) and isinstance(interval2, IndexWithExtent):
                intervals[axis] = interval1.union(interval2)
            else:
                assert (
                    isinstance(interval2, (TileInterval, DomainInterval))
                    and isinstance(interval1, IndexWithExtent)
                ) or (
                    isinstance(interval1, (TileInterval, DomainInterval))
                    and isinstance(interval2, IndexWithExtent)
                )
                intervals[axis] = (
                    interval1
                    if isinstance(interval1, (TileInterval, DomainInterval))
                    else interval2
                )
        return GridSubset(intervals=intervals)


class FieldAccessInfo(eve.Node):
    grid_subset: GridSubset
    global_grid_subset: GridSubset
    dynamic_access: bool = False
    variable_offset_axes: List[Axis] = eve.field(default_factory=list)

    @property
    def is_dynamic(self) -> bool:
        return self.dynamic_access or len(self.variable_offset_axes) > 0

    def axes(self):
        yield from self.grid_subset.axes()

    @property
    def shape(self):
        return self.grid_subset.shape

    @property
    def overapproximated_shape(self):
        return self.grid_subset.overapproximated_shape

    def apply_iteration(self, grid_subset: GridSubset):
        res_intervals = dict(self.grid_subset.intervals)
        for axis, field_interval in self.grid_subset.intervals.items():
            if axis in grid_subset.intervals:
                grid_interval = grid_subset.intervals[axis]
                assert isinstance(field_interval, IndexWithExtent)
                extent = field_interval.extent
                if isinstance(grid_interval, DomainInterval):
                    if axis in self.global_grid_subset.intervals:
                        res_intervals[axis] = self.global_grid_subset.intervals[axis]
                    else:
                        res_intervals[axis] = DomainInterval(
                            start=AxisBound(
                                axis=axis,
                                level=grid_interval.start.level,
                                offset=grid_interval.start.offset + extent[0],
                            ),
                            end=AxisBound(
                                axis=axis,
                                level=grid_interval.end.level,
                                offset=grid_interval.end.offset + extent[1],
                            ),
                        )
                elif isinstance(grid_interval, TileInterval):
                    res_intervals[axis] = TileInterval(
                        axis=axis,
                        tile_size=grid_interval.tile_size,
                        start_offset=grid_interval.start_offset + extent[0],
                        end_offset=grid_interval.end_offset + extent[1],
                        domain_limit=grid_interval.domain_limit,
                    )
                else:
                    assert field_interval.value == grid_interval.value
                    extent = (
                        min(extent) + grid_interval.extent[0],
                        max(extent) + grid_interval.extent[1],
                    )
                    res_intervals[axis] = IndexWithExtent(
                        axis=axis, value=field_interval.value, extent=extent
                    )
        return FieldAccessInfo(
            grid_subset=GridSubset(intervals=res_intervals),
            dynamic_access=self.dynamic_access,
            variable_offset_axes=self.variable_offset_axes,
            global_grid_subset=self.global_grid_subset,
        )

    def union(self, other: "FieldAccessInfo"):
        grid_subset = self.grid_subset.union(other.grid_subset)
        global_subset = self.global_grid_subset.union(other.global_grid_subset)
        variable_offset_axes = [
            axis
            for axis in Axis.dims_3d()
            if axis in self.variable_offset_axes or axis in other.variable_offset_axes
        ]
        return FieldAccessInfo(
            grid_subset=grid_subset,
            dynamic_access=self.dynamic_access or other.dynamic_access,
            variable_offset_axes=variable_offset_axes,
            global_grid_subset=global_subset,
        )

    def clamp_full_axis(self, axis):
        grid_subset = GridSubset(intervals=self.grid_subset.intervals)
        interval = self.grid_subset.intervals[axis]
        full_interval = DomainInterval(
            start=AxisBound(level=common.LevelMarker.START, offset=0, axis=axis),
            end=AxisBound(level=common.LevelMarker.END, offset=0, axis=axis),
        )
        res_interval = DomainInterval.union(
            full_interval,
            self.global_grid_subset.intervals.get(axis, full_interval),
        )
        if isinstance(interval, DomainInterval):
            interval_union = DomainInterval.union(interval, res_interval)
            grid_subset.intervals[axis] = interval_union
        else:
            grid_subset.intervals[axis] = res_interval
            grid_subset = grid_subset.set_interval(axis, res_interval)
        return FieldAccessInfo(
            grid_subset=grid_subset,
            dynamic_access=self.dynamic_access,
            variable_offset_axes=self.variable_offset_axes,
            global_grid_subset=self.global_grid_subset,
        )

    def untile(self, tile_axes: Sequence[Axis]) -> "FieldAccessInfo":
        res_intervals = {}
        for axis, interval in self.grid_subset.intervals.items():
            if isinstance(interval, TileInterval) and axis in tile_axes:
                res_intervals[axis] = self.global_grid_subset.intervals[axis]
            else:
                res_intervals[axis] = interval
        return FieldAccessInfo(
            grid_subset=GridSubset(intervals=res_intervals),
            global_grid_subset=self.global_grid_subset,
            dynamic_access=self.dynamic_access,
            variable_offset_axes=self.variable_offset_axes,
        )

    def restricted_to_index(self, axis: Axis, extent: Tuple[int, int] = (0, 0)):
        return FieldAccessInfo(
            grid_subset=self.grid_subset.restricted_to_index(axis=axis, extent=extent),
            global_grid_subset=self.global_grid_subset,
            dynamic_access=self.dynamic_access,
            variable_offset_axes=self.variable_offset_axes,
        )


class Memlet(eve.Node):
    field: eve.Coerced[eve.SymbolRef]
    access_info: FieldAccessInfo
    connector: eve.Coerced[eve.SymbolRef]
    is_read: bool
    is_write: bool

    def union(self, other):
        assert self.field == other.field
        return Memlet(
            field=self.field,
            access_info=self.access_info.union(other.access_info),
            connector=self.field,
            is_read=self.is_read or other.is_read,
            is_write=self.is_write or other.is_write,
        )

    def remove_read(self):
        return Memlet(
            field=self.field,
            access_info=self.access_info,
            connector=self.connector,
            is_read=False,
            is_write=self.is_write,
        )

    def remove_write(self):
        return Memlet(
            field=self.field,
            access_info=self.access_info,
            connector=self.connector,
            is_read=self.is_read,
            is_write=False,
        )


class Decl(LocNode):
    name: eve.Coerced[eve.SymbolName]
    dtype: common.DataType

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if type(self) is Decl:
            raise TypeError("Trying to instantiate `Decl` abstract class.")
        super().__init__(*args, **kwargs)


class FieldDecl(Decl):
    strides: Tuple[Union[int, str], ...]
    data_dims: Tuple[int, ...] = eve.field(default_factory=tuple)
    access_info: FieldAccessInfo
    storage: StorageType

    @property
    def shape(self):
        access_info = self.access_info
        for axis in self.access_info.variable_offset_axes:
            access_info = access_info.clamp_full_axis(axis)
        ijk_shape = access_info.grid_subset.shape
        return ijk_shape + tuple(self.data_dims)

    def axes(self):
        yield from self.access_info.grid_subset.axes()

    @property
    def is_dynamic(self) -> bool:
        return self.access_info.is_dynamic

    def with_set_access_info(self, access_info: FieldAccessInfo) -> "FieldDecl":
        return FieldDecl(
            name=self.name,
            dtype=self.dtype,
            strides=self.strides,
            data_dims=self.data_dims,
            access_info=access_info,
        )


class Literal(common.Literal, Expr):
    pass


class ScalarAccess(common.ScalarAccess, Expr):
    name: eve.Coerced[eve.SymbolRef]


class VariableKOffset(common.VariableKOffset[Expr]):
    pass


class IndexAccess(common.FieldAccess, Expr):
    offset: Optional[Union[common.CartesianOffset, VariableKOffset]]


class AssignStmt(common.AssignStmt[Union[ScalarAccess, IndexAccess], Expr], Stmt):

    _dtype_validation = common.assign_stmt_dtype_validation(strict=True)


class MaskStmt(Stmt):
    mask: Expr
    body: List[Stmt]

    @datamodels.validator("mask")
    def mask_is_boolean_field_expr(self, attribute: datamodels.Attribute, v: Expr) -> None:
        if v.dtype != common.DataType.BOOL:
            raise ValueError("Mask must be a boolean expression.")


class HorizontalRestriction(common.HorizontalRestriction[Stmt], Stmt):
    pass


class UnaryOp(common.UnaryOp[Expr], Expr):
    pass


class BinaryOp(common.BinaryOp[Expr], Expr):
    _dtype_propagation = common.binary_op_dtype_propagation(strict=True)


class TernaryOp(common.TernaryOp[Expr], Expr):
    _dtype_propagation = common.ternary_op_dtype_propagation(strict=True)


class Cast(common.Cast[Expr], Expr):  # type: ignore
    pass


class NativeFuncCall(common.NativeFuncCall[Expr], Expr):
    _dtype_propagation = common.native_func_call_dtype_propagation(strict=True)


class While(common.While[Stmt, Expr], Stmt):
    pass


class ScalarDecl(Decl):
    pass


class LocalScalarDecl(ScalarDecl):
    pass


class SymbolDecl(ScalarDecl):
    def to_dace_symbol(self):
        return get_dace_symbol(self.name, self.dtype)


class Temporary(FieldDecl):
    pass


class ComputationNode(LocNode):
    # mapping connector names to tuple of field name and access info
    read_memlets: List[Memlet]
    write_memlets: List[Memlet]

    @datamodels.validator("read_memlets")
    @datamodels.validator("write_memlets")
    def _validator(self, attribute: datamodels.Attribute, value: List[Memlet]) -> None:
        conns: Dict[eve.SymbolRef, Set[eve.SymbolRef]] = {}
        for memlet in value:
            conns.setdefault(memlet.field, set())
            if memlet.connector in conns[memlet.field]:
                raise ValueError(f"Found multiple Memlets for connector '{memlet.connector}'")
            conns[memlet.field].add(memlet.connector)

    @property
    def read_fields(self):
        return set(ml.field for ml in self.read_memlets)

    @property
    def write_fields(self):
        return set(ml.field for ml in self.write_memlets)

    @property
    def input_connectors(self):
        return set(ml.connector for ml in self.read_memlets)

    @property
    def output_connectors(self):
        return set(ml.connector for ml in self.write_memlets)


class IterationNode(eve.Node):
    grid_subset: GridSubset


class Tasklet(ComputationNode, IterationNode, eve.SymbolTableTrait):
    decls: List[LocalScalarDecl]
    stmts: List[Stmt]
    grid_subset: GridSubset = GridSubset.single_gridpoint()


class DomainMap(ComputationNode, IterationNode):
    index_ranges: List[Range]
    schedule: MapSchedule
    computations: List[Union[Tasklet, DomainMap, NestedSDFG]]


class ComputationState(IterationNode):
    computations: List[Union[Tasklet, DomainMap]]


class DomainLoop(IterationNode, ComputationNode):
    axis: Axis
    index_range: Range
    loop_states: List[Union[ComputationState, DomainLoop]]


class NestedSDFG(ComputationNode, eve.SymbolTableTrait):
    label: eve.Coerced[eve.SymbolRef]
    field_decls: List[FieldDecl]
    symbol_decls: List[SymbolDecl]
    states: List[Union[DomainLoop, ComputationState]]


# There are circular type references with string placeholders. These statements let datamodels resolve those.
DomainMap.update_forward_refs()  # type: ignore[attr-defined]
DomainLoop.update_forward_refs()  # type: ignore[attr-defined]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import warnings
from typing import List, Set

from gt4py.cartesian.gtc import gtir
from gt4py.eve import NodeVisitor


class DefinitiveAssignmentAnalysis(NodeVisitor):
    """
    Analyse gtir stencil expression for access to undefined symbols.

    A symbol is said to be defined if it was assigned to in the code-path of its usage. In
    other words: If a symbol is defined in both branches of an if-statement it may be used
    outside. If a symbol is only defined in a single branch it may only be used inside that
    branch unless it was already defined previously (as this is equal to assigning to the
    previous value again). Note that whether a symbols is defined is independent of the actual
    result of the condition.
    """

    def visit_IfStmt(self, node: gtir.FieldIfStmt, *, alive_vars: Set[str], **kwargs) -> None:
        true_branch_vars = {*alive_vars}
        false_branch_vars = {*alive_vars}
        self.visit(node.true_branch, alive_vars=true_branch_vars, **kwargs)
        self.visit(node.false_branch, alive_vars=false_branch_vars, **kwargs)
        alive_vars.update(true_branch_vars & false_branch_vars)

    def visit_ParAssignStmt(
        self, node: gtir.ParAssignStmt, *, alive_vars: Set[str], **kwargs
    ) -> None:
        self.visit(node.right, alive_vars=alive_vars, **kwargs)
        alive_vars.add(node.left.name)

    def visit_FieldAccess(
        self,
        node: gtir.FieldAccess,
        *,
        alive_vars: Set[str],
        invalid_accesses: List[gtir.FieldAccess],
        **kwargs,
    ) -> None:
        if node.name not in alive_vars:
            invalid_accesses.append(node)

    @classmethod
    def apply(cls, gtir_stencil_expr: gtir.Stencil) -> List[gtir.FieldAccess]:
        """Execute analysis and return all accesses to undefined symbols."""
        invalid_accesses: List[gtir.FieldAccess] = []
        DefinitiveAssignmentAnalysis().visit(
            gtir_stencil_expr,
            alive_vars=set(gtir_stencil_expr.param_names),
            invalid_accesses=invalid_accesses,
        )
        return invalid_accesses


analyze = DefinitiveAssignmentAnalysis.apply


def check(gtir_stencil_expr: gtir.Stencil) -> gtir.Stencil:
    """Execute definitive assignment analysis and warn on errors."""
    invalid_accesses = analyze(gtir_stencil_expr)
    for invalid_access in invalid_accesses:
        warnings.warn(f"`{invalid_access.name}` may be uninitialized.")

    return gtir_stencil_expr



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import math
import typing
from typing import Any, Dict, Tuple, Union

from gt4py import eve
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.gtc.common import LevelMarker


def _iter_field_names(
    node: Union[gtir.Stencil, gtir.ParAssignStmt]
) -> eve.utils.XIterable[gtir.FieldAccess]:
    return node.walk_values().if_isinstance(gtir.FieldDecl).getattr("name").unique()


class KBoundaryVisitor(eve.NodeVisitor):
    """For every field compute the boundary in k, e.g. (2, -1) if [k_origin-2, k_origin+k_domain-1] is accessed."""

    def visit_Stencil(self, node: gtir.Stencil, **kwargs: Any) -> Dict[str, Tuple[int, int]]:
        field_boundaries = {name: (-math.inf, -math.inf) for name in _iter_field_names(node)}
        for vloop in node.vertical_loops:
            self.generic_visit(vloop.body, vloop=vloop, field_boundaries=field_boundaries, **kwargs)
        # if there is no left or right boundary set to zero
        for name, b in field_boundaries.items():
            field_boundaries[name] = (
                b[0] if b[0] != -math.inf else 0,
                b[1] if b[1] != -math.inf else 0,
            )
        return typing.cast(Dict[str, Tuple[int, int]], field_boundaries)

    def visit_FieldAccess(
        self,
        node: gtir.FieldAccess,
        vloop: gtir.VerticalLoop,
        field_boundaries: Dict[str, Tuple[Union[float, int], Union[float, int]]],
        include_center_interval: bool,
        **kwargs: Any,
    ):
        boundary = field_boundaries[node.name]
        interval = vloop.interval
        if not isinstance(node.offset, gtir.VariableKOffset):
            if interval.start.level == LevelMarker.START and (
                include_center_interval or interval.end.level == LevelMarker.START
            ):
                boundary = (max(-interval.start.offset - node.offset.k, boundary[0]), boundary[1])
            if (
                include_center_interval or interval.start.level == LevelMarker.END
            ) and interval.end.level == LevelMarker.END:
                boundary = (boundary[0], max(interval.end.offset + node.offset.k, boundary[1]))
        if node.name in [decl.name for decl in vloop.temporaries] and (
            boundary[0] > 0 or boundary[1] > 0
        ):
            raise TypeError(f"Invalid access with offset in k to temporary field {node.name}.")
        assert node.name in field_boundaries
        field_boundaries[node.name] = boundary


def compute_k_boundary(
    node: gtir.Stencil, include_center_interval=True
) -> Dict[str, Tuple[int, int]]:
    # loop from START to END is not considered as it might be empty. additional check possible in the future
    return KBoundaryVisitor().visit(node, include_center_interval=include_center_interval)


def compute_min_k_size(node: gtir.Stencil, include_center_interval=True) -> int:
    """Compute the required number of k levels to run a stencil."""
    min_size_start = 0
    min_size_end = 0
    for vloop in node.vertical_loops:
        if vloop.interval.start.level == LevelMarker.START and (
            include_center_interval or vloop.interval.end.level == LevelMarker.START
        ):
            min_size_start = max(min_size_start, vloop.interval.end.offset)
        elif (
            include_center_interval or vloop.interval.start.level == LevelMarker.END
        ) and vloop.interval.end.level == LevelMarker.END:
            min_size_end = max(min_size_end, -vloop.interval.start.offset)
    return min_size_start + min_size_end



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from functools import lru_cache
from typing import TYPE_CHECKING

import dace
import numpy as np

from gt4py import eve
from gt4py.cartesian.gtc import common


if TYPE_CHECKING:
    import gt4py.cartesian.gtc.daceir as dcir


def data_type_to_dace_typeclass(data_type):
    dtype = np.dtype(common.data_type_to_typestr(data_type))
    return dace.dtypes.typeclass(dtype.type)


def get_axis_bound_str(axis_bound, var_name):
    from gt4py.cartesian.gtc.common import LevelMarker

    if axis_bound is None:
        return ""
    elif axis_bound.level == LevelMarker.END:
        return f"{var_name}{axis_bound.offset:+d}"
    else:
        return f"{axis_bound.offset}"


def get_axis_bound_dace_symbol(axis_bound: "dcir.AxisBound"):
    from gt4py.cartesian.gtc.common import LevelMarker

    if axis_bound is None:
        return

    elif axis_bound.level == LevelMarker.END:
        return axis_bound.axis.domain_dace_symbol() + axis_bound.offset
    else:
        return axis_bound.offset


def get_axis_bound_diff_str(axis_bound1, axis_bound2, var_name: str):

    if axis_bound1 <= axis_bound2:
        axis_bound1, axis_bound2 = axis_bound2, axis_bound1
        sign = "-"
    else:
        sign = ""

    if axis_bound1.level != axis_bound2.level:
        var = var_name
    else:
        var = ""
    return f"{sign}({var}{axis_bound1.offset-axis_bound2.offset:+d})"


@lru_cache(maxsize=None)
def get_dace_symbol(name: eve.SymbolRef, dtype: common.DataType = common.DataType.INT32):
    return dace.symbol(name, dtype=data_type_to_dace_typeclass(dtype))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import warnings
from typing import Any

from gt4py.cartesian.gtc import common, oir
from gt4py.eve import NodeTranslator


class AdjacentLoopMerging(NodeTranslator):
    @staticmethod
    def _mergeable(a: oir.VerticalLoop, b: oir.VerticalLoop) -> bool:
        if a.loop_order != b.loop_order:
            return False
        a_interval = a.sections[-1].interval
        b_interval = b.sections[0].interval
        if a.loop_order == common.LoopOrder.BACKWARD:
            a_lim = a_interval.start
            b_lim = b_interval.end
        else:
            a_lim = a_interval.end
            b_lim = b_interval.start
        return a_lim.level == b_lim.level and a_lim.offset == b_lim.offset

    @staticmethod
    def _merge(a: oir.VerticalLoop, b: oir.VerticalLoop) -> oir.VerticalLoop:
        sections = a.sections + b.sections
        if a.caches or b.caches:
            warnings.warn("AdjacentLoopMerging pass removed previously declared caches")
        return oir.VerticalLoop(
            loop_order=a.loop_order,
            sections=sections,
            caches=[],
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        if not node.vertical_loops:
            return self.generic_visit(node, **kwargs)
        vertical_loops = [self.visit(node.vertical_loops[0], **kwargs)]
        for vertical_loop in node.vertical_loops[1:]:
            vertical_loop = self.visit(vertical_loop, **kwargs)
            mergeable = self._mergeable(vertical_loops[-1], vertical_loop)
            if mergeable:
                vertical_loops[-1] = self._merge(vertical_loops[-1], vertical_loop)
            else:
                vertical_loops.append(vertical_loop)

        return oir.Stencil(
            name=node.name,
            params=node.params,
            vertical_loops=vertical_loops,
            declarations=node.declarations,
            loc=node.loc,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, Dict

from gt4py import eve
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.gtc.common import DataType, GTCPostconditionError


class _GTIRResolveAuto(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """
    Replaces AUTO dtype by a concrete dtype.

    Note that currently only temporaries (FieldDecl/FieldAccess) can have AUTO type.

    Precondition: All dtype are set, but some can be set to AUTO
    Postcondition: All dtypes are concrete (no AUTO)
    """

    class _GTIRUpdateAutoDecl(eve.NodeTranslator):
        """Updates FieldDecls with resolved types."""

        def visit_FieldDecl(
            self, node: gtir.FieldDecl, new_symbols: Dict[str, Any], **kwargs: Any
        ) -> gtir.FieldDecl:
            if node.dtype == DataType.AUTO:
                dtype = new_symbols[node.name].dtype
                return gtir.FieldDecl(
                    name=node.name, dtype=dtype, dimensions=node.dimensions, loc=node.loc
                )
            else:
                return node

    def visit_FieldAccess(
        self, node: gtir.FieldAccess, *, symtable: Dict[str, Any], **kwargs: Any
    ) -> gtir.FieldAccess:
        if symtable[node.name].dtype == DataType.AUTO:
            assert "new_dtype" in kwargs
            symtable[node.name].dtype = kwargs["new_dtype"]
        return gtir.FieldAccess(
            name=node.name,
            offset=self.visit(node.offset, symtable=symtable, **kwargs),
            data_index=self.visit(node.data_index, symtable=symtable, **kwargs),
            dtype=symtable[node.name].dtype,
            loc=node.loc,
        )

    def visit_ParAssignStmt(self, node: gtir.ParAssignStmt, **kwargs: Any) -> gtir.ParAssignStmt:
        right = self.visit(node.right, **kwargs)
        left = self.visit(node.left, new_dtype=right.dtype, **kwargs)
        return gtir.ParAssignStmt(left=left, right=right, loc=node.loc)

    def visit_Stencil(self, node: gtir.Stencil, **kwargs: Any) -> gtir.Stencil:
        result = self.generic_visit(node, **kwargs)
        result = self._GTIRUpdateAutoDecl().visit(result, new_symbols=kwargs["symtable"])

        if not all(
            result.walk_values()
            .if_hasattr("dtype")
            .getattr("dtype")
            .map(lambda x: x not in {DataType.AUTO, DataType.INVALID, DataType.DEFAULT})
        ):
            raise GTCPostconditionError(expected="No AUTO, INVALID or DEFAULT dtype in tree.")

        return result


class _GTIRPropagateDtypeToAccess(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """
    Propagates dtype from Decl to Access.

    Precondition: Decls have dtype (not None), can be AUTO or DEFAULT
    Postcondition: All dtypes of Access are not None
    """

    def visit_FieldAccess(self, node: gtir.FieldAccess, **kwargs: Any) -> gtir.FieldAccess:
        return gtir.FieldAccess(
            name=node.name,
            offset=self.visit(node.offset, **kwargs),
            data_index=self.visit(node.data_index, **kwargs),
            dtype=kwargs["symtable"][node.name].dtype,
            loc=node.loc,
        )

    def visit_ScalarAccess(
        self, node: gtir.ScalarAccess, *, symtable: Dict[str, Any], **kwargs: Any
    ) -> gtir.ScalarAccess:
        return gtir.ScalarAccess(name=node.name, dtype=symtable[node.name].dtype, loc=node.loc)


def resolve_dtype(node: gtir.Stencil) -> gtir.Stencil:
    return _GTIRResolveAuto().visit(_GTIRPropagateDtypeToAccess().visit(node))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import List, Optional, Tuple, Union

from gt4py import eve
from gt4py.cartesian.gtc import common
from gt4py.cartesian.gtc.definitions import Extent
from gt4py.eve import datamodels


# --- Misc ---
class AxisName(eve.StrEnum):
    I = "I"  # noqa: E741 (ambiguous variable name)
    J = "J"
    K = "K"


# NOTE HorizontalMask in npir differs from common.HorizontalMask:
# - They are expressed relative to the iteration domain of the statement
# - Each axis is a tuple of two common.AxisBound instead of common.HorizontalInterval
class HorizontalMask(eve.Node):
    i: Tuple[common.AxisBound, common.AxisBound]
    j: Tuple[common.AxisBound, common.AxisBound]


# --- Decls ---
@eve.utils.noninstantiable
class Decl(eve.Node):
    name: eve.Coerced[eve.SymbolName]
    dtype: common.DataType


class ScalarDecl(Decl):
    """Scalar per grid point.

    Used for API scalar parameters. Local scalars never have data_dims.

    """

    pass


class LocalScalarDecl(Decl):
    """Scalar per grid point.

    Used for API scalar parameters. Local scalars never have data_dims.

    """

    pass


class FieldDecl(Decl):
    """General field shared across HorizontalBlocks."""

    dimensions: Tuple[bool, bool, bool]
    data_dims: Tuple[int, ...] = eve.field(default_factory=tuple)
    extent: Extent


class TemporaryDecl(Decl):
    """
    Temporary field shared across HorizontalBlocks.

    Parameters
    ----------
    offset: Origin of the temporary field.
    padding: Buffer added to compute domain as field size.

    """

    data_dims: Tuple[int, ...] = eve.field(default_factory=tuple)
    offset: Tuple[int, int]
    padding: Tuple[int, int]


# --- Expressions ---
@eve.utils.noninstantiable
class Expr(common.Expr):
    pass


@eve.utils.noninstantiable
class VectorLValue(Expr):
    pass


class ScalarLiteral(common.Literal, Expr):
    kind = common.ExprKind.SCALAR

    @datamodels.validator("dtype")
    def is_defined(self, attribute: datamodels.Attribute, dtype: common.DataType) -> None:
        undefined = [common.DataType.AUTO, common.DataType.DEFAULT, common.DataType.INVALID]
        if dtype in undefined:
            raise ValueError("npir.Literal may not have undefined data type.")


class ScalarCast(common.Cast[Expr], Expr):
    kind = common.ExprKind.SCALAR


class VectorCast(common.Cast[Expr], Expr):
    kind = common.ExprKind.FIELD


class Broadcast(Expr):
    expr: Expr
    kind: common.ExprKind = common.ExprKind.FIELD


class VarKOffset(common.VariableKOffset[Expr]):
    pass


class FieldSlice(VectorLValue):
    name: eve.Coerced[eve.SymbolRef]
    i_offset: int
    j_offset: int
    k_offset: Union[int, VarKOffset]
    data_index: List[Expr] = eve.field(default_factory=list)
    kind: common.ExprKind = common.ExprKind.FIELD

    @datamodels.validator("data_index")
    def data_indices_are_scalar(
        self, attribute: datamodels.Attribute, data_index: List[Expr]
    ) -> None:
        for index in data_index:
            if index.kind != common.ExprKind.SCALAR:
                raise ValueError("Data indices must be scalars")


class ParamAccess(Expr):
    name: eve.Coerced[eve.SymbolRef]
    kind: common.ExprKind = common.ExprKind.SCALAR


class LocalScalarAccess(VectorLValue):
    name: eve.Coerced[eve.SymbolRef]
    kind: common.ExprKind = common.ExprKind.FIELD


class VectorArithmetic(common.BinaryOp[Expr], Expr):
    op: Union[common.ArithmeticOperator, common.ComparisonOperator]

    _dtype_propagation = common.binary_op_dtype_propagation(strict=True)


class VectorLogic(common.BinaryOp[Expr], Expr):
    op: common.LogicalOperator


class VectorUnaryOp(common.UnaryOp[Expr], Expr):
    pass


class VectorTernaryOp(common.TernaryOp[Expr], Expr):
    _dtype_propagation = common.ternary_op_dtype_propagation(strict=True)


class NativeFuncCall(common.NativeFuncCall[Expr], Expr):
    _dtype_propagation = common.native_func_call_dtype_propagation(strict=True)


# --- Statements ---
@eve.utils.noninstantiable
class Stmt(common.Stmt):
    pass


class VectorAssign(common.AssignStmt[VectorLValue, Expr], Stmt):
    # NOTE HorizontalMask in npir differs from common.HorizontalMask (see above)
    horizontal_mask: Optional[HorizontalMask] = None

    @datamodels.validator("right")
    def right_is_field_kind(self, attribute: datamodels.Attribute, right: Expr) -> None:
        if right.kind != common.ExprKind.FIELD:
            raise ValueError("right is not a common.ExprKind.FIELD")

    _dtype_validation = common.assign_stmt_dtype_validation(strict=True)


class While(common.While[Stmt, Expr], Stmt):
    pass


# --- Control Flow ---
class HorizontalBlock(common.LocNode, eve.SymbolTableTrait):
    body: List[Stmt]
    extent: Extent
    declarations: List[LocalScalarDecl]


class VerticalPass(common.LocNode):
    body: List[HorizontalBlock]
    lower: common.AxisBound
    upper: common.AxisBound
    direction: common.LoopOrder


class Computation(common.LocNode, eve.SymbolTableTrait):
    arguments: List[str]
    api_field_decls: List[FieldDecl]
    param_decls: List[ScalarDecl]
    temp_decls: List[TemporaryDecl]
    vertical_passes: List[VerticalPass]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Implementation of stencil functions defined with symbolic Python functions.

This module contains functions to generate callable objects implementing
a high-level stencil function definition using a specific code generating backend.
"""

import types
from typing import TYPE_CHECKING, Any, Dict, Type

from gt4py.cartesian import backend as gt_backend
from gt4py.cartesian import frontend as gt_frontend
from gt4py.cartesian.stencil_builder import StencilBuilder
from gt4py.cartesian.type_hints import StencilFunc


if TYPE_CHECKING:
    from gt4py.cartesian.definitions import BuildOptions
    from gt4py.cartesian.stencil_object import StencilObject


def load_stencil(
    frontend_name: str,
    backend_name: str,
    definition_func: StencilFunc,
    externals: Dict[str, Any],
    build_options: "BuildOptions",
) -> Type["StencilObject"]:
    """Generate a new class object implementing the provided definition."""
    # Load components
    backend_cls = gt_backend.from_name(backend_name)
    if backend_cls is None:
        raise ValueError(f"Unknown backend name ({backend_name})")

    frontend = gt_frontend.from_name(frontend_name)
    if frontend is None:
        raise ValueError(f"Invalid frontend name ({frontend_name})")

    builder = StencilBuilder(
        definition_func, options=build_options, backend=backend_name, frontend=frontend
    ).with_externals(externals)

    return builder.build()


def gtscript_loader(
    definition_func: StencilFunc,
    backend: str,
    build_options: "BuildOptions",
    externals: Dict[str, Any],
) -> "StencilObject":
    if not isinstance(definition_func, types.FunctionType):
        raise ValueError("Invalid stencil definition object ({obj})".format(obj=definition_func))

    if not build_options.name:
        build_options.name = f"{definition_func.__name__}"
    stencil_class = load_stencil("gtscript", backend, definition_func, externals, build_options)

    return stencil_class()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import pytest

from gt4py import eve
from gt4py.eve.extended_typing import Any, ClassVar, List

from .. import definitions


@pytest.fixture
def symtable_node_and_expected_symbols():
    node = definitions.make_node_with_symbol_table()
    symbols = {
        node.node_with_name.name: node.node_with_name,
        node.node_with_default_name.name: node.node_with_default_name,
        node.compound_with_name.node_with_name.name: node.compound_with_name.node_with_name,
    }
    symbols.update({n.name: n for n in node.list_with_name})

    yield node, symbols


class _NodeWithSymbolName(eve.Node):
    name: eve.Coerced[eve.SymbolName]


class _NodeWithSymbolTable(eve.Node, eve.SymbolTableTrait):
    symbols: List[_NodeWithSymbolName]


@pytest.fixture
def node_with_duplicated_names_maker():
    def _maker():
        return _NodeWithSymbolTable(
            symbols=[_NodeWithSymbolName(name="repeated"), _NodeWithSymbolName(name="repeated")]
        )

    yield _maker


class TestSymbolTable:
    def test_no_duplicated_names(self, node_with_duplicated_names_maker):
        with pytest.raises(ValueError, match="Multiple definitions of symbol"):
            node_with_duplicated_names_maker()

    def test_symbol_table_creation(self, symtable_node_and_expected_symbols):
        node, expected_symbols = symtable_node_and_expected_symbols
        collected_symtable = node.annex.symtable
        assert isinstance(node.annex.symtable, dict)
        assert all(isinstance(key, str) for key in collected_symtable)

    def test_symbol_table_collection(self, symtable_node_and_expected_symbols):
        node, expected_symbols = symtable_node_and_expected_symbols
        collected_symtable = node.annex.symtable
        assert collected_symtable == expected_symbols
        assert all(
            collected_symtable[symbol_name] is symbol_node
            for symbol_name, symbol_node in expected_symbols.items()
        )

    def test_extra_node_symbols(self):
        class NodeWithName(eve.Node):
            name: eve.Coerced[eve.SymbolName]

        class NodeWithRef(eve.Node):
            ref_name: eve.Coerced[eve.SymbolRef]

        class NodeWithSymbolTable(eve.Node, eve.traits.ValidatedSymbolTableTrait):
            symbols: List[NodeWithRef]

            _NODE_SYMBOLS_: ClassVar[List] = []

        NodeWithSymbolTable.update_forward_refs(locals())

        with pytest.raises(ValueError, match="'foo'"):
            NodeWithSymbolTable(symbols=[NodeWithRef(ref_name="foo")])

        NodeWithSymbolTable._NODE_SYMBOLS_.append(NodeWithName(name="foo"))
        NodeWithSymbolTable(symbols=[NodeWithRef(ref_name=eve.SymbolRef("foo"))])


def test_visitor_with_symbol_table_trait(node_with_symbol_table):
    class BareVisitor(eve.visitors.NodeVisitor):
        def visit_Node(self, node: eve.concepts.RootNode, **kwargs: Any) -> Any:
            assert "symtable" in kwargs

    with pytest.raises(AssertionError, match="'symtable'"):
        BareVisitor().visit(node_with_symbol_table)

    class ExtendedVisitor(eve.traits.VisitorWithSymbolTableTrait):
        def visit_Node(self, node: eve.concepts.RootNode, **kwargs: Any) -> Any:
            assert "symtable" in kwargs
            assert "symbol_name" in kwargs["symtable"]

    ExtendedVisitor().visit(node_with_symbol_table)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import dace
from dace.transformation.dataflow import TrivialMapElimination
from dace.transformation.helpers import nest_state_subgraph
from dace.transformation.interstate import InlineTransients


class NoEmptyEdgeTrivialMapElimination(TrivialMapElimination):
    """Eliminate trivial maps like TrivialMapElimination, with additional conditions in can_be_applied."""

    def can_be_applied(self, graph, expr_index, sdfg, permissive=False):

        if not super().can_be_applied(graph, expr_index, sdfg, permissive=permissive):
            return False

        map_entry = self.map_entry
        map_exit = graph.exit_node(map_entry)
        if map_entry.map.schedule not in {
            dace.ScheduleType.Sequential,
            dace.ScheduleType.CPU_Multicore,
        }:
            return False
        if any(
            edge.data.is_empty() for edge in (graph.in_edges(map_entry) + graph.out_edges(map_exit))
        ):
            return False
        return True


class InlineThreadLocalTransients(dace.transformation.SingleStateTransformation):
    """
    Inline and tile thread-local transients.

    Inlines transients like `dace.transformations.interstate.InlineTransients`, however only applies to OpenMP map
    scopes but also makes the resulting local arrays persistent and thread-local. This reproduces `cpu_kfirst`-style
    transient tiling.
    """

    map_entry = dace.transformation.transformation.PatternNode(dace.nodes.MapEntry)

    @classmethod
    def expressions(cls):
        return [dace.sdfg.utils.node_path_graph(cls.map_entry)]

    def can_be_applied(self, graph, expr_index, sdfg, permissive=False):
        map_entry = self.map_entry

        if not map_entry.schedule == dace.ScheduleType.CPU_Multicore:
            return False

        scope_subgraph = graph.scope_subgraph(map_entry, include_entry=False, include_exit=False)
        if len(scope_subgraph) > 1 or not isinstance(
            scope_subgraph.nodes()[0], dace.nodes.NestedSDFG
        ):
            return False

        candidates = InlineTransients._candidates(sdfg, graph, scope_subgraph.nodes()[0])
        return len(candidates) > 0

    def apply(self, graph, sdfg):
        map_entry = self.map_entry

        scope_subgraph = graph.scope_subgraph(map_entry, include_entry=False, include_exit=False)
        nsdfg_node = scope_subgraph.nodes()[0]
        candidates = InlineTransients._candidates(sdfg, graph, nsdfg_node)
        InlineTransients.apply_to(sdfg, nsdfg=nsdfg_node, save=False)
        for name in candidates:
            if name in sdfg.arrays:
                continue
            array: dace.data.Array = nsdfg_node.sdfg.arrays[name]
            shape = [dace.symbolic.overapproximate(s) for s in array.shape]
            strides = [1]
            total_size = shape[0]
            for s in reversed(shape[1:]):
                strides = [s * strides[0], *strides]
                total_size *= s
            array.shape = shape
            array.strides = strides
            array.total_size = total_size
            array.storage = dace.StorageType.CPU_ThreadLocal
            array.lifetime = dace.AllocationLifetime.Persistent


def nest_sequential_map_scopes(sdfg: dace.SDFG):
    """Nest map scopes of sequential maps.

    Nest scope subgraphs of sequential maps in NestedSDFG's to force eagerly offsetting pointers on each iteration, to
    avoid more complex pointer arithmetic on each Tasklet's invocation.
    This is performed in an inner-map-first order to avoid revisiting the graph after changes.
    """

    def _process_map(sdfg: dace.SDFG, state: dace.SDFGState, map_entry: dace.nodes.MapEntry):
        for node in state.scope_children()[map_entry]:
            if isinstance(node, dace.nodes.NestedSDFG):
                nest_sequential_map_scopes(node.sdfg)
            elif isinstance(node, dace.nodes.MapEntry):
                _process_map(sdfg, state, node)
        if map_entry.schedule == dace.ScheduleType.Sequential:
            subgraph = state.scope_subgraph(map_entry, include_entry=False, include_exit=False)
            nest_state_subgraph(sdfg, state, subgraph)

    state: dace.SDFGState
    for state in sdfg.nodes():
        for map_entry in filter(
            lambda n: isinstance(n, dace.nodes.MapEntry), state.scope_children()[None]
        ):
            _process_map(sdfg, state, map_entry)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from gt4py import eve
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.definitions import Extent

from .utils import (
    AccessCollector,
    collect_symbol_names,
    compute_horizontal_block_extents,
    symbol_name_creator,
)


class HorizontalExecutionMerging(eve.NodeTranslator):
    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        all_names = collect_symbol_names(node)
        return self.generic_visit(
            node,
            block_extents=compute_horizontal_block_extents(node),
            new_symbol_name=symbol_name_creator(all_names),
            **kwargs,
        )

    def visit_VerticalLoopSection(
        self,
        node: oir.VerticalLoopSection,
        *,
        block_extents: Dict[int, Extent],
        new_symbol_name: Callable[[str], str],
        **kwargs: Any,
    ) -> oir.VerticalLoopSection:
        @dataclass
        class UncheckedHorizontalExecution:
            # local replacement without type checking for type-checked oir node
            # required to reach reasonable run times for large node counts
            body: List[oir.Stmt]
            declarations: List[oir.LocalScalar]
            loc: Optional[eve.SourceLocation]

            assert set(oir.HorizontalExecution.__datamodel_fields__.keys()) == {
                "loc",
                "body",
                "declarations",
            }, (
                "Unexpected field in oir.HorizontalExecution, "
                "probably UncheckedHorizontalExecution needs an update"
            )

            @classmethod
            def from_oir(cls, hexec: oir.HorizontalExecution):
                return cls(body=hexec.body, declarations=hexec.declarations, loc=hexec.loc)

            def to_oir(self) -> oir.HorizontalExecution:
                return oir.HorizontalExecution(
                    body=self.body, declarations=self.declarations, loc=self.loc
                )

        horizontal_executions = [
            UncheckedHorizontalExecution.from_oir(node.horizontal_executions[0])
        ]
        new_block_extents = [block_extents[id(node.horizontal_executions[0])]]
        last_writes = AccessCollector.apply(node.horizontal_executions[0]).write_fields()

        for this_hexec in node.horizontal_executions[1:]:
            last_extent = new_block_extents[-1]

            this_offset_reads = {
                name
                for name, offsets in AccessCollector.apply(this_hexec).read_offsets().items()
                if any(off[0] != 0 or off[1] != 0 for off in offsets)
            }

            reads_with_offset_after_write = last_writes & this_offset_reads
            this_extent = block_extents[id(this_hexec)]

            if reads_with_offset_after_write or last_extent != this_extent:
                # Cannot merge: simply append to list
                horizontal_executions.append(UncheckedHorizontalExecution.from_oir(this_hexec))
                new_block_extents.append(this_extent)
                last_writes = AccessCollector.apply(this_hexec).write_fields()
            else:
                # Merge
                duplicated_locals = {
                    decl.name for decl in horizontal_executions[-1].declarations
                } & {decl.name for decl in this_hexec.declarations}
                # Map from old to new scalar names applied to the second horizontal execution
                scalar_map = {name: new_symbol_name(name) for name in duplicated_locals}
                locals_symtable = {decl.name: decl for decl in this_hexec.declarations}

                new_body = self.visit(this_hexec.body, scalar_map=scalar_map, **kwargs)

                this_not_duplicated = [
                    decl for decl in this_hexec.declarations if decl.name not in duplicated_locals
                ]
                this_mapped = [
                    oir.LocalScalar(name=scalar_map[name], dtype=locals_symtable[name].dtype)
                    for name in duplicated_locals
                ]

                horizontal_executions[-1] = UncheckedHorizontalExecution(
                    body=horizontal_executions[-1].body + new_body,
                    declarations=(
                        horizontal_executions[-1].declarations + this_not_duplicated + this_mapped
                    ),
                    loc=horizontal_executions[-1].loc,
                )
                last_writes |= AccessCollector.apply(new_body).write_fields()

        return oir.VerticalLoopSection(
            interval=node.interval,
            horizontal_executions=[hexec.to_oir() for hexec in horizontal_executions],
            loc=node.loc,
        )

    def visit_ScalarAccess(
        self, node: oir.ScalarAccess, *, scalar_map: Dict[str, str], **kwargs: Any
    ) -> oir.ScalarAccess:
        return oir.ScalarAccess(
            name=scalar_map[node.name] if node.name in scalar_map else node.name,
            dtype=node.dtype,
            loc=node.loc,
        )


@dataclass
class OnTheFlyMerging(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    """Merges consecutive horizontal executions inside parallel vertical loops by introducing redundant computations.

    Limitations:
    * Works on the level of whole horizontal executions, no full dependency analysis is performed (common subexpression and dead code eliminitation at a later stage can work around this limitation).
    * The chosen default merge limits are totally arbitrary.
    """

    max_horizontal_execution_body_size: int = 100
    allow_expensive_function_duplication: bool = False

    def visit_CartesianOffset(
        self,
        node: common.CartesianOffset,
        *,
        shift: Optional[Tuple[int, int, int]] = None,
        **kwargs: Any,
    ) -> common.CartesianOffset:
        if shift:
            di, dj, dk = shift
            return common.CartesianOffset(i=node.i + di, j=node.j + dj, k=node.k + dk)
        return self.generic_visit(node, **kwargs)

    def visit_FieldAccess(
        self,
        node: oir.FieldAccess,
        *,
        offset_symbol_map: Dict[Tuple[str, Tuple[int, int, int]], str] = None,
        **kwargs: Any,
    ) -> Union[oir.FieldAccess, oir.ScalarAccess]:
        if offset_symbol_map:
            offset = self.visit(node.offset, **kwargs)
            key = node.name, (offset.i, offset.j, offset.k)
            if key in offset_symbol_map:
                return oir.ScalarAccess(name=offset_symbol_map[key], dtype=node.dtype, loc=node.loc)
        return self.generic_visit(node, **kwargs)

    def visit_ScalarAccess(
        self, node: oir.ScalarAccess, *, scalar_map: Dict[str, str], **kwargs: Any
    ) -> oir.ScalarAccess:
        return oir.ScalarAccess(
            name=scalar_map[node.name] if node.name in scalar_map else node.name,
            dtype=node.dtype,
            loc=node.loc,
        )

    def _merge(
        self,
        horizontal_executions: List[oir.HorizontalExecution],
        symtable: Dict[str, Any],
        new_symbol_name: Callable[[str], str],
        protected_fields: Set[str],
    ) -> List[oir.HorizontalExecution]:
        """Recursively merge horizontal executions.

        Uses the following algorithm:
        1. Get output fields of the first horizontal execution.
        2. Check in which following h. execs. the outputs are read.
        3. Duplicate the body of the first h. exec. for each read access (with corresponding offset) and prepend it to the depending h. execs.
        4. Recurse on the resulting h. execs.
        """
        if len(horizontal_executions) <= 1:
            return horizontal_executions
        first, *others = horizontal_executions
        first_accesses = AccessCollector.apply(first)
        other_accesses = AccessCollector.apply(others)

        def first_fields_rewritten_later() -> bool:
            return bool(first_accesses.fields() & other_accesses.write_fields())

        def first_has_large_body() -> bool:
            return len(first.body) > self.max_horizontal_execution_body_size

        def first_writes_protected() -> bool:
            return bool(protected_fields & first_accesses.write_fields())

        def first_has_expensive_function_call() -> bool:
            if self.allow_expensive_function_duplication:
                return False
            nf = common.NativeFunction
            expensive_calls = {
                nf.SIN,
                nf.COS,
                nf.TAN,
                nf.ARCSIN,
                nf.ARCCOS,
                nf.ARCTAN,
                nf.SINH,
                nf.COSH,
                nf.TANH,
                nf.ARCSINH,
                nf.ARCCOSH,
                nf.ARCTANH,
                nf.SQRT,
                nf.EXP,
                nf.LOG,
                nf.GAMMA,
                nf.CBRT,
            }
            calls = first.walk_values().if_isinstance(oir.NativeFuncCall).getattr("func")
            return any(call in expensive_calls for call in calls)

        def first_has_variable_access() -> bool:
            return first_accesses.has_variable_access()

        def first_has_horizontal_restriction() -> bool:
            return any(first.walk_values().if_isinstance(oir.HorizontalRestriction))

        if (
            first_fields_rewritten_later()
            or first_writes_protected()
            or first_has_large_body()
            or first_has_expensive_function_call()
            or first_has_variable_access()
            or first_has_horizontal_restriction()
        ):
            return [first] + self._merge(others, symtable, new_symbol_name, protected_fields)

        first_scalars = {decl.name for decl in first.declarations}
        writes = first_accesses.write_fields()
        others_otf = []
        for horizontal_execution in others:
            read_offsets: Set[Tuple[int, int, int]] = set()
            read_offsets = read_offsets.union(
                *(
                    offsets
                    for field, offsets in AccessCollector.apply(horizontal_execution)
                    .cartesian_accesses()
                    .read_offsets()
                    .items()
                    if field in writes
                )
            )

            if not read_offsets:
                others_otf.append(horizontal_execution)
                continue

            duplicated_locals = first_scalars & {
                decl.name for decl in horizontal_execution.declarations
            }
            scalar_map = {name: new_symbol_name(name) for name in duplicated_locals}

            offset_symbol_map = {
                (name, o): new_symbol_name(name) for name in writes for o in read_offsets
            }

            # 4 contributions to the new declarations list
            combined_symtable = {**symtable, **first.annex.symtable}
            decls_from_later = [
                d for d in horizontal_execution.declarations if d.name not in duplicated_locals
            ]
            decls_from_first = [
                d
                for d in first.declarations
                if d not in horizontal_execution.declarations or d.name in duplicated_locals
            ]
            decls_renamed_locals_in_later = [
                oir.LocalScalar(name=new_name, dtype=combined_symtable[old_name].dtype)
                for old_name, new_name in scalar_map.items()
            ]
            new_decls = [
                oir.LocalScalar(name=new_name, dtype=combined_symtable[old_name].dtype)
                for (old_name, _), new_name in offset_symbol_map.items()
            ]

            declarations = (
                decls_from_later + decls_from_first + decls_renamed_locals_in_later + new_decls
            )

            merged = oir.HorizontalExecution(
                body=self.visit(
                    horizontal_execution.body,
                    offset_symbol_map=offset_symbol_map,
                    scalar_map=scalar_map,
                ),
                declarations=declarations,
                loc=first.loc,
            )
            for offset in read_offsets:
                merged.body = (
                    self.visit(
                        first.body,
                        shift=offset,
                        offset_symbol_map=offset_symbol_map,
                        scalar_map={},
                    )
                    + merged.body
                )
            others_otf.append(merged)

        return self._merge(others_otf, symtable, new_symbol_name, protected_fields)

    def visit_VerticalLoopSection(
        self, node: oir.VerticalLoopSection, **kwargs: Any
    ) -> oir.VerticalLoopSection:

        last_vls = None
        next_vls = node
        applied = True
        while applied:
            last_vls = next_vls
            next_vls = oir.VerticalLoopSection(
                interval=last_vls.interval,
                horizontal_executions=self._merge(last_vls.horizontal_executions, **kwargs),
                loc=node.loc,
            )
            applied = len(next_vls.horizontal_executions) < len(last_vls.horizontal_executions)

        return next_vls

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> oir.VerticalLoop:
        if node.loop_order != common.LoopOrder.PARALLEL:
            return node
        sections = self.visit(node.sections, **kwargs)
        accessed = AccessCollector.apply(sections).fields()
        return oir.VerticalLoop(
            loop_order=node.loop_order,
            sections=sections,
            caches=[c for c in node.caches if c.name in accessed],
            loc=node.loc,
        )

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> oir.Stencil:
        protected_fields = set(str(n.name) for n in node.params)
        new_symbol_name = symbol_name_creator(collect_symbol_names(node))
        vertical_loops = []
        for vl in reversed(node.vertical_loops):
            vl = self.visit(
                vl, new_symbol_name=new_symbol_name, protected_fields=protected_fields, **kwargs
            )
            vertical_loops.append(vl)
            protected_fields |= AccessCollector.apply(vl).read_fields()
        vertical_loops = list(reversed(vertical_loops))

        accessed = AccessCollector.apply(vertical_loops).fields()
        return oir.Stencil(
            name=node.name,
            params=node.params,
            vertical_loops=vertical_loops,
            declarations=[d for d in node.declarations if d.name in accessed],
            loc=node.loc,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import collections
import dataclasses
import enum
import itertools
import numbers
from typing import Any, Callable, Optional, Sequence, Tuple

import hypothesis.strategies as hyp_st
import numpy as np
from hypothesis.extra import numpy as hyp_np


# ---- Test Suites utilities ----
class SymbolKind(enum.Enum):
    NONE = 0
    GLOBAL_STRATEGY = 1
    GLOBAL_SET = 2
    SINGLETON = 3
    PARAMETER = 4
    FIELD = 5


@dataclasses.dataclass(frozen=True)
class _SymbolStrategy:
    kind: SymbolKind
    boundary: Optional[Sequence[Tuple[int, int]]]
    axes: Optional[str]
    data_dims: Optional[Tuple[int, ...]]
    value_st_factory: Callable[..., hyp_st.SearchStrategy]


@dataclasses.dataclass(frozen=True)
class _SymbolValueTuple:
    kind: str
    boundary: Sequence[Tuple[int, int]]
    values: Tuple[Any]


def global_name(*, singleton=None, symbol=None, one_of=None, in_range=None):
    """Define a *global* symbol."""
    if singleton is not None:
        assert symbol is None and one_of is None and in_range is None
        return _SymbolValueTuple(kind=SymbolKind.SINGLETON, boundary=None, values=(singleton,))

    if symbol is not None:
        assert singleton is None and one_of is None and in_range is None
        return _SymbolValueTuple(kind=SymbolKind.GLOBAL_SET, boundary=None, values=(symbol,))

    if one_of is not None:
        assert (
            singleton is None
            and symbol is None
            and in_range is None
            and isinstance(one_of, collections.abc.Sequence)
        )
        return _SymbolValueTuple(kind=SymbolKind.GLOBAL_SET, boundary=None, values=one_of)

    if in_range is not None:
        assert singleton is None and symbol is None and one_of is None and len(in_range) == 2
        return _SymbolStrategy(
            kind=SymbolKind.GLOBAL_STRATEGY,
            boundary=None,
            axes=None,
            data_dims=None,
            value_st_factory=lambda dt: scalar_value_st(dt, in_range[0], in_range[1]),
        )

    else:
        raise AssertionError("Missing value descriptor")


def field(*, in_range, boundary=None, axes=None, data_dims=None, extent=None):
    """Define a *field* symbol."""
    assert (boundary is not None or extent is not None) and len(in_range) == 2
    boundary = boundary or [(abs(e[0]), abs(e[1])) for e in extent]
    extent = extent or [(-b[0], b[1]) for b in boundary]
    assert all((-b[0], b[1]) == (e[0], e[1]) for b, e in zip(boundary, extent))
    assert all((b[0] >= 0 and b[1]) >= 0 for b in boundary)

    return _SymbolStrategy(
        kind=SymbolKind.FIELD,
        boundary=boundary,
        axes=axes,
        data_dims=data_dims or tuple(),
        value_st_factory=lambda dt: scalar_value_st(dt, in_range[0], in_range[1]),
    )


def parameter(*, one_of=None, in_range=None):
    """Define a *parameter* symbol."""
    if one_of is not None:
        assert in_range is None and isinstance(one_of, collections.abc.Sequence)
        return _SymbolStrategy(
            kind=SymbolKind.PARAMETER,
            boundary=None,
            axes=None,
            data_dims=None,
            value_st_factory=lambda dt: one_of_values_st(one_of).map(dt),
        )

    elif in_range is not None:
        assert one_of is None and len(in_range) == 2
        return _SymbolStrategy(
            kind=SymbolKind.PARAMETER,
            boundary=None,
            axes=None,
            data_dims=None,
            value_st_factory=lambda dt: scalar_value_st(dt, in_range[0], in_range[1]),
        )

    else:
        raise AssertionError("Missing value descriptor")


def none():
    """Define the symbol ``None``."""
    return _SymbolStrategy(
        kind=SymbolKind.NONE,
        boundary=None,
        axes=None,
        data_dims=None,
        value_st_factory=lambda dt: hyp_st.none(),
    )


# ---- Custom Hypothesis strategies ----
def scalar_value_st(dtype, min_value, max_value, allow_nan=False):
    """Hypothesis strategy for `dtype` scalar values in range [min_value, max_value]."""
    allow_infinity = not (np.isfinite(min_value) and np.isfinite(max_value))

    if issubclass(dtype.type, numbers.Real):
        value_st = hyp_st.floats(
            min_value,
            max_value,
            allow_infinity=allow_infinity,
            allow_nan=allow_nan,
            width=dtype.itemsize * 8,
        )
    elif issubclass(dtype.type, numbers.Integral):
        value_st = hyp_st.integers(min_value, max_value)

    return value_st.map(dtype.type)


def one_of_values_st(args):
    """Hypothesis strategy returning one of the values passed in the arguments."""
    if len(args) == 0:
        return hyp_st.just(None)
    else:
        return hyp_st.sampled_from(args)


def ndarray_shape_st(sizes):
    """Hypothesis strategy for shapes of `ndims` dimensions and size within `size_range`."""
    return hyp_st.tuples(*[hyp_st.integers(min_size, max_size) for (min_size, max_size) in sizes])


def derived_shape_st(shape_st, extra: Sequence[Optional[int]]):
    """Hypothesis strategy for extending shapes generated from a provided strategy with some extra padding.

    If an element of extra contains None, the item will be dropped from the final shape, otherwise,
    both shape and extra elements are summed together.
    """
    return hyp_st.builds(
        lambda shape: tuple(
            [d + e for d, e in itertools.zip_longest(shape, extra, fillvalue=0) if e is not None]
        ),
        shape_st,
    )


def ndarray_in_range_st(dtype, shape_st, value_range):
    """Hypothesis strategy for ndarrays generated using the provided `dtype` and `shape` strategies and the `value_range`."""
    return hyp_np.arrays(
        dtype=dtype,
        shape=shape_st,
        elements=scalar_value_st(dtype, min_value=value_range[0], max_value=value_range[1]),
        fill=scalar_value_st(dtype, min_value=value_range[0], max_value=value_range[1]),
    )


def ndarray_st(dtype, shape_strategy, value_st_factory):
    """Hypothesis strategy for ndarrays generated using the provided `dtype` and `shape` and `value_st_factory` strategies/factories."""
    return hyp_np.arrays(
        dtype=dtype,
        shape=shape_strategy,
        elements=value_st_factory(dtype),
        fill=value_st_factory(dtype),
    )


# ---- Utility functions ----
@hyp_st.composite
def draw_from_strategies(draw, strategies):
    """Generate a Hypothesis strategy example by composing named strategies from a dictionary.

    Parameters
    ----------
    draw : `function`
        Hypothesis :func:`draw` implementation.
    strategies : `dict`
        Named strategies.
        - ``name``: Hypothesis strategy (`function`).

    Returns
    ----------
    args: `dict`
        A dictionary of Hypothesis examples.
        - ``name``: Hypothesis strategy example (``value``).
    """
    args = dict()
    for name, strategy in strategies.items():
        args[name] = draw(strategy)

    return args


def composite_strategy_factory(dtypes, strategy_factories):
    strategy_dict = dict()
    for name, factory in strategy_factories.items():
        strategy_dict[name] = factory(dtypes[name])

    @hyp_st.composite
    def strategy(draw):
        args = dict()
        for name, strategy in strategy_dict.items():
            args[name] = draw(strategy)
        return args

    return strategy


def composite_implementation_strategy_factory(
    dtypes, validation_strategy_factories, global_boundaries
):
    """Generate strategy for run-time :class:`StencilTestSuite` parameters."""
    strategy_dict = dict()
    for name, factory in validation_strategy_factories.items():
        strategy_dict[name] = factory(dtypes[name])

    @hyp_st.composite
    def strategy(draw):
        inputs = dict()
        for name, strat in strategy_dict.items():
            inputs[name] = draw(strat)

        return inputs, dict()

    return strategy



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, Optional, Set, Type

import pytest

import gt4py.eve.codegen

from .test_utils import name_with_cases  # noqa: F401


# -- Name tests --
def test_name(name_with_cases):  # noqa: F811  # pytest fixture not detected
    name = gt4py.eve.codegen.Name(name_with_cases.pop("words"))
    for case, cased_string in name_with_cases.items():
        assert name.as_case(case) == cased_string
        for other_case, other_cased_string in name_with_cases.items():
            if other_case == gt4py.eve.utils.CaseStyleConverter.CASE_STYLE.CONCATENATED:
                with pytest.raises(ValueError, match="split a simply concatenated"):
                    other_name = gt4py.eve.codegen.Name.from_string(other_cased_string, other_case)
                    assert other_name.as_case(case) == cased_string
            else:
                other_name = gt4py.eve.codegen.Name.from_string(other_cased_string, other_case)
                assert other_name.as_case(case) == cased_string


# -- Template tests --
def fmt_tpl_maker(skeleton, keys, valid=True):
    if valid:
        transformed_keys = {k: "{{{key}}}".format(key=k) for k in keys}
        return gt4py.eve.codegen.FormatTemplate(skeleton.format(**transformed_keys))
    else:
        return None


def string_tpl_maker(skeleton, keys, valid=True):
    if valid:
        transformed_keys = {k: "${}".format(k) for k in keys}
        return gt4py.eve.codegen.StringTemplate(skeleton.format(**transformed_keys))
    else:
        return None


def jinja_tpl_maker(skeleton, keys, valid=True):
    if valid:
        transformed_keys = {k: "{{{{ {} }}}}".format(k) for k in keys}
    else:
        transformed_keys = {k: "{{{{ --%{} }}}}".format(k) for k in keys}
    return gt4py.eve.codegen.JinjaTemplate(skeleton.format(**transformed_keys))


def mako_tpl_maker(skeleton, keys, valid=True):
    if valid:
        transformed_keys = {k: "${{{}}}".format(k) for k in keys}
    else:
        transformed_keys = {k: "${{$<%$${}}}".format(k) for k in keys}
    return gt4py.eve.codegen.MakoTemplate(skeleton.format(**transformed_keys))


@pytest.fixture(params=[fmt_tpl_maker, string_tpl_maker, jinja_tpl_maker, mako_tpl_maker])
def template_maker(
    request,
) -> Callable[[str, Set[str], bool], Optional[gt4py.eve.codegen.Template]]:
    return request.param


def test_template_definition(template_maker):
    skeleton = "aaa {s} bbbb {i} cccc"
    data = {"s": "STRING", "i": 1}

    template_maker(skeleton, data.keys())
    with pytest.raises(gt4py.eve.codegen.TemplateDefinitionError):
        t = template_maker(skeleton, data.keys(), False)
        if t is None:
            # Some template engines do not check templates at definition
            raise gt4py.eve.codegen.TemplateDefinitionError


def test_template_rendering(template_maker):
    skeleton = "aaa {s} bbbb {i} cccc"
    data = {"s": "STRING", "i": 1}
    template = template_maker(skeleton, data.keys())
    assert template.render(**data) == "aaa STRING bbbb 1 cccc"
    assert template.render(data, i=2) == "aaa STRING bbbb 2 cccc"

    with pytest.raises(gt4py.eve.codegen.TemplateRenderingError):
        template.render()


# -- TemplatedGenerator tests --
class _BaseTestGenerator(gt4py.eve.codegen.TemplatedGenerator):
    KEYWORDS = ("BASE", "ONE")

    def visit_IntKind(self, node, **kwargs):
        return f"ONE INTKIND({node.value})"

    def visit_SourceLocation(self, node, **kwargs):
        return f"SourceLocation<line:{node.line}, column:{node.column}, source: {node.source}>"

    LocationNode = gt4py.eve.codegen.FormatTemplate("LocationNode {{{loc}}}")

    SimpleNode = gt4py.eve.codegen.JinjaTemplate(
        "|{{ bool_value }}, {{ int_value }}, {{ float_value }}, {{ str_value }}, {{ bytes_value }}, "
        "{{ int_kind }}, {{ _this_node.str_kind.__class__.__name__ }}|"
    )

    def visit_SimpleNode(self, node, **kwargs):
        return f"SimpleNode {{{self.generic_visit(node, **kwargs)}}}"

    CompoundNode = gt4py.eve.codegen.MakoTemplate(
        """
----CompoundNode [BASE]----
    - location: ${location}
    - simple: ${simple}
    - simple_opt: <has_optionals ? (${_this_node.simple_opt.int_value is not None}, ${_this_node.simple_opt.float_value is not None}, ${_this_node.simple_opt.str_value is not None})>
    - other_simple_opt: <is_present ? ${_this_node.other_simple_opt is not None}>
"""
    )

    def visit_CompoundNode(self, node, **kwargs):
        return "TemplatedGenerator result:\n" + self.generic_visit(node, **kwargs)


class _InheritedTestGenerator(_BaseTestGenerator):
    KEYWORDS = ("INHERITED", "OTHER")

    def visit_IntKind(self, node, **kwargs):
        return f"OTHER INTKIND({node.value})"

    CompoundNode = gt4py.eve.codegen.MakoTemplate(
        """
----CompoundNode [INHERITED]----
    - location: ${location}
    - simple: ${simple}
    - simple_opt: <has_optionals ? (${_this_node.simple_opt.int_value is not None}, ${_this_node.simple_opt.float_value is not None}, ${_this_node.simple_opt.str_value is not None})>
    - other_simple_opt: <is_present ? ${_this_node.other_simple_opt is not None}>
"""
    )


class _FaultyTestGenerator1(gt4py.eve.codegen.TemplatedGenerator):
    CompoundNode = gt4py.eve.codegen.FormatTemplate(
        """
Line
Another {MISSING_loc}"""
    )


class _FaultyTestGenerator2(gt4py.eve.codegen.TemplatedGenerator):
    CompoundNode = gt4py.eve.codegen.StringTemplate(
        """
Line
Another $f{ffloc"""
    )


class _FaultyTestGenerator3(gt4py.eve.codegen.TemplatedGenerator):
    CompoundNode = gt4py.eve.codegen.JinjaTemplate(
        """
|{{ bool_value }}, {{ MISSING_int_value }}, {{ float_value }}, {{ str_value }}, {{ bytes_value }},
        {{ int_kind }}, {{ _this_node.str_kind.__class__.WRONG__name__ }}|
    """
    )


class _FaultyTestGenerator4(gt4py.eve.codegen.TemplatedGenerator):
    CompoundNode = gt4py.eve.codegen.MakoTemplate(
        """
----CompoundNode [BASE]----
    - location: ${location}
    - simple: ${MISSING_simple}
    - simple_opt: <has_optionals ? (${_this_node.simple_opt.int_value is not None}, ${_this_node.simple_opt.float_value is not None}, ${_this_node.simple_opt.str_value is not None})>
    - other_simple_opt: <is_present ? ${_this_node.other_simple_opt is not None}>
"""
    )


@pytest.fixture(params=[_BaseTestGenerator, _InheritedTestGenerator])
def templated_generator(request) -> Type[gt4py.eve.codegen.TemplatedGenerator]:
    return request.param


@pytest.fixture(
    params=[
        _FaultyTestGenerator1,
        _FaultyTestGenerator2,
        _FaultyTestGenerator3,
        _FaultyTestGenerator4,
    ]
)
def faulty_templated_generator(request) -> Type[gt4py.eve.codegen.TemplatedGenerator]:
    return request.param


def test_templated_generator(templated_generator, fixed_compound_node):
    rendered_code = templated_generator.apply(fixed_compound_node)
    assert rendered_code.find("TemplatedGenerator result:\n") >= 0
    assert rendered_code.find("----CompoundNode [") >= 0
    assert rendered_code.find("LocationNode {") >= 0
    assert rendered_code.find("SimpleNode {|") >= 0
    assert rendered_code.find("<has_optionals ? (True, True, False)>") >= 0
    assert rendered_code.find("<is_present ? False>") >= 0

    for keyword in templated_generator.KEYWORDS:
        assert rendered_code.find(keyword) >= 0


def test_templated_generator_exceptions(faulty_templated_generator, fixed_compound_node):
    with pytest.raises(gt4py.eve.codegen.TemplateRenderingError, match="when rendering node"):
        faulty_templated_generator.apply(fixed_compound_node)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2022, ETH Zurich
# All rights reserved.
#
# This file is part of the GT4Py project and the GridTools framework.
# GT4Py is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or any later
# version. See the LICENSE.txt file at the top-level directory of this
# distribution for a copy of the license or check <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import abstractmethod
from typing import Callable, Optional, Protocol, Sequence, Type, Union

from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.passes.oir_optimizations.caches import (
    IJCacheDetection,
    KCacheDetection,
    PruneKCacheFills,
    PruneKCacheFlushes,
)
from gt4py.cartesian.gtc.passes.oir_optimizations.horizontal_execution_merging import (
    HorizontalExecutionMerging,
    OnTheFlyMerging,
)
from gt4py.cartesian.gtc.passes.oir_optimizations.inlining import MaskInlining
from gt4py.cartesian.gtc.passes.oir_optimizations.mask_stmt_merging import MaskStmtMerging
from gt4py.cartesian.gtc.passes.oir_optimizations.pruning import (
    NoFieldAccessPruning,
    UnreachableStmtPruning,
)
from gt4py.cartesian.gtc.passes.oir_optimizations.temporaries import (
    LocalTemporariesToScalars,
    WriteBeforeReadTemporariesToScalars,
)
from gt4py.cartesian.gtc.passes.oir_optimizations.vertical_loop_merging import AdjacentLoopMerging
from gt4py.eve.visitors import NodeVisitor


PassT = Union[Callable[[oir.Stencil], oir.Stencil], Type[NodeVisitor]]


class OirPipeline(Protocol):
    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError("Missing implementation of __hash__")

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError("Missing implementation of __repr__")

    @abstractmethod
    def run(self, oir: oir.Stencil) -> oir.Stencil:
        raise NotImplementedError("Missing implementation of run")


class DefaultPipeline(OirPipeline):
    """
    OIR passes pipeline runs passes in order and allows skipping.

    May only call existing passes and may not contain any pass logic itself.
    """

    def __init__(
        self, *, skip: Optional[Sequence[PassT]] = None, add_steps: Optional[Sequence[PassT]] = None
    ):
        self.skip = list(skip or [])
        self.add_steps = list(add_steps or [])

    @staticmethod
    def all_steps() -> Sequence[PassT]:
        return [
            AdjacentLoopMerging,
            HorizontalExecutionMerging,
            OnTheFlyMerging,
            LocalTemporariesToScalars,
            WriteBeforeReadTemporariesToScalars,
            MaskStmtMerging,
            MaskInlining,
            UnreachableStmtPruning,
            NoFieldAccessPruning,
            IJCacheDetection,
            KCacheDetection,
            PruneKCacheFills,
            PruneKCacheFlushes,
        ]

    @property
    def steps(self) -> Sequence[PassT]:
        return [step for step in self.all_steps() if step not in self.skip] + self.add_steps

    def __hash__(self) -> int:
        return hash(repr(self))

    def __repr__(self) -> str:
        return str([step.__name__ for step in self.steps])

    def __eq__(self, other):
        return isinstance(other, DefaultPipeline) and self.skip == other.skip

    def run(self, oir: oir.Stencil) -> oir.Stencil:
        for step in self.steps:
            if isinstance(step, type) and issubclass(step, NodeVisitor):
                oir = step().visit(oir)
            else:
                oir = step(oir)
        return oir
